"""Parameter schemas for image-to-video models."""

# Schema dictionary organized by model name
SCHEMAS = {
    "wan2.1-vace-14b": {
        "required": {
            "prompt": {"type": str, "description": "Text prompt to guide generation"},
            "image": {"type": str, "description": "Path or URL to input image"}
        },
        "optional": {
            "negative_prompt": {"type": str, "default": "", "description": "The prompt or prompts not to guide the image generation. Ignored when guidance_scale is less than 1."},
            "video": {"type": list, "description": "The input video (List[PIL.Image.Image]) to be used as a starting point for the generation."},
            "mask": {"type": list, "description": "The input mask (List[PIL.Image.Image]) that defines which video regions to condition on (black) and which to generate (white)."},
            "reference_images": {"type": list, "description": "A list of one or more reference images (List[PIL.Image.Image]) as extra conditioning for the generation."},
            "conditioning_scale": {"type": float, "default": 1.0, "description": "The scale applied to the control conditioning latent stream. Can be a float, List[float], or torch.Tensor."},
            "height": {"type": int, "default": 480, "description": "The height in pixels of the generated video."},
            "width": {"type": int, "default": 832, "description": "The width in pixels of the generated video."},
            "num_frames": {"type": int, "default": 81, "description": "The number of frames in the generated video."},
            "num_inference_steps": {"type": int, "default": 50, "description": "The number of denoising steps. More steps usually lead to higher quality at the expense of slower inference."},
            "guidance_scale": {"type": float, "default": 5.0, "description": "Guidance scale for classifier-free diffusion. Higher values encourage  to be closely linked to the text prompt."},
            "num_videos_per_prompt": {"type": int, "default": 1, "description": "The number of videos to generate per prompt."},
            "generator": {"type": "torch.Generator", "description": "A torch.Generator or List[torch.Generator] to make generation deterministic."},
            "latents": {"type": "torch.Tensor", "description": "Pre-generated noisy latents to be used as inputs for generation."},
            "prompt_embeds": {"type": "torch.Tensor", "description": "Pre-generated text embeddings. Can be used to easily tweak text inputs."},
            "output_type": {"type": str, "default": "np", "description": "The output format of the generated image. Choose between 'PIL.Image' or 'np.array'."},
            "return_dict": {"type": bool, "default": True, "description": "Whether or not to return a WanPipelineOutput instead of a plain tuple."},
            "attention_kwargs": {"type": dict, "description": "A kwargs dictionary passed along to the AttentionProcessor."},
            "callback_on_step_end": {"type": "Callable", "description": "A function called at the end of each denoising step during inference."},
            "callback_on_step_end_tensor_inputs": {"type": list, "description": "The list of tensor inputs for the callback_on_step_end function."},
            "max_sequence_length": {"type": int, "default": 512, "description": "The maximum sequence length of the text encoder."},
            "flow_shift": {"type": float, "default": 5.0, "description": "a value that estimates motion between two frames. A larger flow shift focuses on high motion or transformation. A smaller flow shift focuses on stability."}
        }
    },
    "wan2.1-i2v-14b-720p": {
        "required": {
            "prompt": {"type": str, "description": "Text prompt to guide generation"},
            "image": {"type": str, "description": "Path or URL to input image"}
        },
        "optional": {
            # "prompt": {"type": str, "description": "The prompt or prompts to guide the image generation. If not defined, prompt_embeds must be passed."},
            "negative_prompt": {"type": str, "description": "The prompt or prompts not to guide the image generation. Ignored if guidance_scale is less than 1."},
            "height": {"type": int, "default": 480, "description": "The height of the generated video in pixels."},
            "width": {"type": int, "default": 832, "description": "The width of the generated video in pixels."},
            "num_frames": {"type": int, "default": 81, "description": "The number of frames in the generated video."},
            "num_inference_steps": {"type": int, "default": 50, "description": "The number of denoising steps. More steps can improve quality but are slower."},
            "guidance_scale": {"type": float, "default": 5.0, "description": "Classifier-Free Diffusion guidance scale. Higher values align the video more closely with the prompt."},
            "num_videos_per_prompt": {"type": int, "default": 1, "description": "The number of videos to generate for each prompt."},
            "generator": {"type": "torch.Generator", "description": "A torch.Generator or List[torch.Generator] to make generation deterministic."},
            "latents": {"type": "torch.Tensor", "description": "Pre-generated noisy latents to be used as inputs for generation."},
            "prompt_embeds": {"type": "torch.Tensor", "description": "Pre-generated text embeddings, used as an alternative to the 'prompt' argument."},
            "negative_prompt_embeds": {"type": "torch.Tensor", "description": "Pre-generated negative text embeddings, used as an alternative to the 'negative_prompt' argument."},
            "image_embeds": {"type": "torch.Tensor", "description": "Pre-generated image embeddings, used as an alternative to the 'image' argument."},
            "output_type": {"type": str, "default": "np", "description": "The output format of the generated video. Choose between 'PIL.Image' or 'np.array'."},
            "return_dict": {"type": bool, "default": True, "description": "Whether to return a WanPipelineOutput object instead of a plain tuple."},
            "attention_kwargs": {"type": dict, "description": "A kwargs dictionary passed to the AttentionProcessor."},
            "callback_on_step_end": {"type": "Callable", "description": "A function called at the end of each denoising step during inference."},
            "callback_on_step_end_tensor_inputs": {"type": list, "description": "The list of tensor inputs for the callback_on_step_end function."},
            "max_sequence_length": {"type": int, "default": 512, "description": "The maximum sequence length for the text encoder."}
        }
    },
    "cogvideox-5b-image-to-video": {
        "required": {
            "prompt": {"type": str, "description": "Text prompt to guide generation"},
            "image": {"type": str, "description": "Path or URL to input image"}
        },
        "optional": {
            "negative_prompt": {"type": str, "description": "The prompt or prompts not to guide video generation. Ignored if guidance_scale is less than 1."},
            "height": {"type": int, "default": 480, "description": "The height in pixels of the generated video."},
            "width": {"type": int, "default": 720, "description": "The width in pixels of the generated video."},
            "num_frames": {"type": int, "default": 48, "description": "Number of frames to generate."},
            "num_inference_steps": {"type": int, "default": 50, "description": "The number of denoising steps. More steps can improve quality but are slower."},
            "timesteps": {"type": list, "description": "Custom timesteps to use for the denoising process, must be in descending order."},
            "guidance_scale": {"type": float, "default": 7.0, "description": "Classifier-Free Diffusion guidance scale. Higher values align the video more closely with the prompt."},
            "num_videos_per_prompt": {"type": int, "default": 1, "description": "The number of videos to generate for each prompt."},
            "generator": {"type": "torch.Generator", "description": "A torch.Generator or List[torch.Generator] to make generation deterministic."},
            "latents": {"type": "torch.FloatTensor", "description": "Pre-generated noisy latents to be used as inputs for generation."},
            "prompt_embeds": {"type": "torch.FloatTensor", "description": "Pre-generated text embeddings, used as an alternative to the 'prompt' argument."},
            "negative_prompt_embeds": {"type": "torch.FloatTensor", "description": "Pre-generated negative text embeddings, used as an alternative to the 'negative_prompt' argument."},
            "output_type": {"type": str, "default": "pil", "description": "The output format of the generated video. Choose between 'pil' or 'np.array'."},
            "return_dict": {"type": bool, "default": True, "description": "Whether to return a StableDiffusionXLPipelineOutput object instead of a plain tuple."},
            "attention_kwargs": {"type": dict, "description": "A kwargs dictionary passed to the AttentionProcessor."},
            "callback_on_step_end": {"type": "Callable", "description": "A function called at the end of each denoising step during inference."},
            "callback_on_step_end_tensor_inputs": {"type": list, "description": "The list of tensor inputs for the callback_on_step_end function."},
            "max_sequence_length": {"type": int, "default": 226, "description": "Maximum sequence length in the encoded prompt."}
        }
    },
    "framepack-i2v-hy": {
        "required": {
            # "image": {"type": "Image", "description": "The starting image for the video generation. Accepts PIL.Image, np.ndarray, or torch.Tensor."},
            "prompt": {"type": str, "description": "Text prompt to guide generation"},
            "image": {"type": str, "description": "Path or URL to input image"}
        },
        "optional": {
            # "last_image": {"type": "Image", "description": "The optional ending image for the video generation, useful for image-to-image transitions."},
            # "prompt": {"type": str, "description": "The prompt to guide video generation."},
            "prompt_2": {"type": str, "description": "A secondary prompt for the second text encoder; defaults to the main prompt if not provided."},
            "negative_prompt": {"type": str, "description": "The prompt to avoid during video generation."},
            "negative_prompt_2": {"type": str, "description": "A secondary negative prompt for the second text encoder."},
            "height": {"type": int, "default": 720, "description": "The height in pixels of the generated video."},
            "width": {"type": int, "default": 1280, "description": "The width in pixels of the generated video."},
            "num_frames": {"type": int, "default": 129, "description": "The number of frames in the generated video."},
            "num_inference_steps": {"type": int, "default": 50, "description": "The number of denoising steps."},
            "sigmas": {"type": list, "description": "Custom sigmas for the denoising scheduler."},
            "true_cfg_scale": {"type": float, "default": 1.0, "description": "Enables true classifier-free guidance when > 1.0."},
            "guidance_scale": {"type": float, "default": 6.0, "description": "Guidance scale to control how closely the video adheres to the prompt."},
            "num_videos_per_prompt": {"type": int, "default": 1, "description": "The number of videos to generate per prompt."},
            "generator": {"type": "torch.Generator", "description": "A torch.Generator to make generation deterministic."},
            "image_latents": {"type": "torch.Tensor", "description": "Pre-encoded image latents, bypassing the VAE for the first image."},
            "last_image_latents": {"type": "torch.Tensor", "description": "Pre-encoded image latents, bypassing the VAE for the last image."},
            "prompt_embeds": {"type": "torch.Tensor", "description": "Pre-generated text embeddings, an alternative to 'prompt'."},
            "pooled_prompt_embeds": {"type": "torch.FloatTensor", "description": "Pre-generated pooled text embeddings."},
            "negative_prompt_embeds": {"type": "torch.FloatTensor", "description": "Pre-generated negative text embeddings, an alternative to 'negative_prompt'."},
            "negative_pooled_prompt_embeds": {"type": "torch.FloatTensor", "description": "Pre-generated negative pooled text embeddings."},
            "output_type": {"type": str, "default": "pil", "description": "The output format of the generated video frames ('pil' or 'np.array')."},
            "return_dict": {"type": bool, "default": True, "description": "Whether to return a HunyuanVideoFramepackPipelineOutput object instead of a plain tuple."},
            "attention_kwargs": {"type": dict, "description": "A kwargs dictionary passed to the AttentionProcessor."},
            "clip_skip": {"type": int, "description": "Number of final layers to skip from the CLIP model."},
            "callback_on_step_end": {"type": "Callable", "description": "A function called at the end of each denoising step."},
            "callback_on_step_end_tensor_inputs": {"type": list, "description": "The list of tensor inputs for the callback_on_step_end function."}
        }
    },
    "ltx-video": {
        "required": {
            "prompt": {"type": str, "description": "Text prompt to guide generation"},
            "image": {"type": str, "description": "Path or URL to input image"}
        },
        "optional": {
            "negative_prompt": {"type": str, "description": "The prompt to avoid during video generation."},
            "height": {"type": int, "default": 512, "description": "The height in pixels of the generated video."},
            "width": {"type": int, "default": 704, "description": "The width in pixels of the generated video."},
            "num_frames": {"type": int, "default": 161, "description": "The number of video frames to generate."},
            "num_inference_steps": {"type": int, "default": 50, "description": "The number of denoising steps."},
            "timesteps": {"type": list, "description": "Custom timesteps for the denoising process in descending order."},
            "guidance_scale": {"type": float, "default": 3.0, "description": "Scale for classifier-free guidance."},
            "guidance_rescale": {"type": float, "default": 0.0, "description": "Guidance rescale factor to prevent overexposure."},
            "num_videos_per_prompt": {"type": int, "default": 1, "description": "The number of videos to generate per prompt."},
            "generator": {"type": "torch.Generator", "description": "A torch.Generator to make generation deterministic."},
            "latents": {"type": "torch.Tensor", "description": "Pre-generated noisy latents."},
            "prompt_embeds": {"type": "torch.Tensor", "description": "Pre-generated text embeddings, an alternative to 'prompt'."},
            "prompt_attention_mask": {"type": "torch.Tensor", "description": "Pre-generated attention mask for text embeddings."},
            "negative_prompt_embeds": {"type": "torch.FloatTensor", "description": "Pre-generated negative text embeddings."},
            "negative_prompt_attention_mask": {"type": "torch.FloatTensor", "description": "Pre-generated attention mask for negative text embeddings."},
            "decode_timestep": {"type": float, "default": 0.0, "description": "The timestep at which the generated video is decoded."},
            "decode_noise_scale": {"type": float, "default": None, "description": "Interpolation factor between random noise and denoised latents at decode time."},
            "output_type": {"type": str, "default": "pil", "description": "The output format of the generated video frames ('pil' or 'np.array')."},
            "return_dict": {"type": bool, "default": True, "description": "Whether to return an LTXPipelineOutput object instead of a plain tuple."},
            "attention_kwargs": {"type": dict, "description": "A kwargs dictionary passed to the AttentionProcessor."},
            "callback_on_step_end": {"type": "Callable", "description": "A function called at the end of each denoising step."},
            "callback_on_step_end_tensor_inputs": {"type": list, "description": "The list of tensor inputs for the callback_on_step_end function."},
            "max_sequence_length": {"type": int, "default": 128, "description": "Maximum sequence length for the prompt."}
        }
    },
    "veo-2.0-generate-001": {
        "required": {
            "prompt": {"type": str, "description": "Text prompt to guide generation"},
            "image": {"type": str, "description": "Path or URL to input image"}
        },
        "optional": {
            # "DURATION": {"type": int, "default": 8, "description": "The length of video files that you want to generate. Accepted integer values are 5-8."},
            "negativePrompt": {"type": str, "default": "", "description": "Text string that describes anything you want to discourage the model from generating"},
            "aspectRatio": {"type": str, "default": "16:9", "description": "Defines the aspect ratio of the generated videos. Accepts '16:9' (landscape) or '9:16' (portrait)."},
            "personGeneration": {"type": str, "default": "allow_adult", "description": "Controls whether people or face generation is allowed. Accepts 'allow_adult' or 'disallow'."},
            # "RESOLUTION": {"type": str, "default": "720p", "description": "The resolution of the generated video (Veo 3 models only). Accepts '720p' or '1080p'."},
            "numberOfVideos": {"type": int, "default": 1, "description": "The number of output videos requested, from 1 to 2."},
            "durationSeconds": {"type": int, "default": 8, "description": "Veo 2 only. Length of each output video in seconds, between 5 and 8"},

            # "SEED_NUMBER": {"type": int, "description": "A number from 0 to 4,294,967,295 to make video generation deterministic."}
        }
    },
    "pusa-v1": {
        "required": {
            "prompt": {"type": str, "description": "Text prompt to guide generation"},
            "image": {"type": str, "description": "Path or URL to input image"}
        },
        "optional": {
            "negative_prompt": {"type": str, "description": "The prompt or prompts not to guide video generation. Ignored if guidance_scale is less than 1."},
            "cond_position": {"type": str, "default": "0", "description": "Comma-separated list of frame indices for conditioning. You can use any position from 0 to 20."},
            "noise_multipliers": {"type": str, "default": "0.0", "description": "Comma-separated noise multipliers for conditioning frames. A value of 0 means the condition image is used as totally clean, higher value means add more noise. For I2V, you can use 0.2 or any from 0 to 1. For Start-End-Frame, you can use 0.2,0.4, or any from 0 to 1."},
            "lora_alpha": {"type": float, "default": 1.0, "description": "A bigger alpha would bring more temporal consistency (i.e., make generated frames more like conditioning part), but may also cause small motion or even collapse. We recommend using a value around 1 to 2."},
            "num_inference_steps": {"type": int, "default": 30},
            "num_frames": {"type": int, "default": 81},
        }
    },
    "wan2.1-vace-14b-i2v-fusionx": {
        "required": {
            "prompt": {"type": str, "description": "Text prompt to guide generation"},
            "image": {"type": str, "description": "Path or URL to input image"}
        },
        "optional": {
            "negative_prompt": {"type": str, "default": "", "description": "The prompt or prompts not to guide the image generation. Ignored when guidance_scale is less than 1."},
            "video": {"type": list, "description": "The input video (List[PIL.Image.Image]) to be used as a starting point for the generation."},
            "mask": {"type": list, "description": "The input mask (List[PIL.Image.Image]) that defines which video regions to condition on (black) and which to generate (white)."},
            "reference_images": {"type": list, "description": "A list of one or more reference images (List[PIL.Image.Image]) as extra conditioning for the generation."},
            "conditioning_scale": {"type": float, "default": 1.0, "description": "The scale applied to the control conditioning latent stream. Can be a float, List[float], or torch.Tensor."},
            "height": {"type": int, "default": 480, "description": "The height in pixels of the generated video."},
            "width": {"type": int, "default": 832, "description": "The width in pixels of the generated video."},
            "num_frames": {"type": int, "default": 81, "description": "The number of frames in the generated video."},
            "num_inference_steps": {"type": int, "default": 50, "description": "The number of denoising steps. More steps usually lead to higher quality at the expense of slower inference."},
            "guidance_scale": {"type": float, "default": 5.0, "description": "Guidance scale for classifier-free diffusion. Higher values encourage  to be closely linked to the text prompt."},
            "num_videos_per_prompt": {"type": int, "default": 1, "description": "The number of videos to generate per prompt."},
            "generator": {"type": "torch.Generator", "description": "A torch.Generator or List[torch.Generator] to make generation deterministic."},
            "latents": {"type": "torch.Tensor", "description": "Pre-generated noisy latents to be used as inputs for generation."},
            "prompt_embeds": {"type": "torch.Tensor", "description": "Pre-generated text embeddings. Can be used to easily tweak text inputs."},
            "output_type": {"type": str, "default": "np", "description": "The output format of the generated image. Choose between 'PIL.Image' or 'np.array'."},
            "return_dict": {"type": bool, "default": True, "description": "Whether or not to return a WanPipelineOutput instead of a plain tuple."},
            "attention_kwargs": {"type": dict, "description": "A kwargs dictionary passed along to the AttentionProcessor."},
            "callback_on_step_end": {"type": "Callable", "description": "A function called at the end of each denoising step during inference."},
            "callback_on_step_end_tensor_inputs": {"type": list, "description": "The list of tensor inputs for the callback_on_step_end function."},
            "max_sequence_length": {"type": int, "default": 512, "description": "The maximum sequence length of the text encoder."},
            "flow_shift": {"type": float, "default": 5.0, "description": "a value that estimates motion between two frames. A larger flow shift focuses on high motion or transformation. A smaller flow shift focuses on stability."}
        }
    },
    "luma-ray-2": {
        "required": {
            "prompt": {"type": str, "description": "Text prompt to guide generation"},
            "image": {"type": str, "description": "URL to input image"}
        },
        "optional": {}
    }

    # Add more models as needed
}


from .constants import FeatureType, SUPPORTED_FEATURE_TYPES, is_valid_feature_type, get_feature_types

__all__ = [
    "SUPPORTED_FEATURE_TYPES",
    "FeatureType",
    "get_feature_types",
    "is_valid_feature_type"
]


"""Parameter schemas for text-to-video models."""

# Schema dictionary organized by model name
SCHEMAS = {
    "wan2.1-t2v-14b": {
        "required": {
            "prompt": {"type": str, "description": "Text prompt to guide generation"},
        },
        "optional": {
            "negative_prompt": {"type": str, "description": "The prompt or prompts to avoid during image generation. Ignored when guidance_scale is less than 1."},
            "height": {"type": int, "default": 480, "description": "The height in pixels of the generated image."},
            "width": {"type": int, "default": 832, "description": "The width in pixels of the generated image."},
            "num_frames": {"type": int, "default": 81, "description": "The number of frames in the generated video."},
            "num_inference_steps": {"type": int, "default": 50, "description": "The number of denoising steps. More steps usually lead to higher quality at the expense of slower inference."},
            "guidance_scale": {"type": float, "default": 5.0, "description": "Guidance scale for classifier-free diffusion. Higher values encourage generation to be closely linked to the text prompt."},
            "num_videos_per_prompt": {"type": int, "default": 1, "description": "The number of videos to generate per prompt."},
            "generator": {"type": "torch.Generator", "description": "A torch.Generator or List[torch.Generator] to make generation deterministic."},
            "latents": {"type": "torch.Tensor", "description": "Pre-generated noisy latents to be used as inputs for image generation."},
            "prompt_embeds": {"type": "torch.Tensor", "description": "Pre-generated text embeddings, used as an alternative to the 'prompt' argument."},
            "output_type": {"type": str, "default": "np", "description": "The output format of the generated image. Choose between 'PIL.Image' or 'np.array'."},
            "return_dict": {"type": bool, "default": True, "description": "Whether or not to return a WanPipelineOutput object instead of a plain tuple."},
            "attention_kwargs": {"type": dict, "description": "A kwargs dictionary passed to the AttentionProcessor."},
            "callback_on_step_end": {"type": "Callable", "description": "A function called at the end of each denoising step during inference."},
            "callback_on_step_end_tensor_inputs": {"type": list, "description": "The list of tensor inputs for the callback_on_step_end function."},
            "max_sequence_length": {"type": int, "default": 512, "description": "The maximum sequence length of the text encoder."}
        }
    },
    "wan2.1-14b-t2v-fusionx": {
        "required": {
            "prompt": {"type": str, "description": "Text prompt to guide generation"},
        },
        "optional": {
            "negative_prompt": {"type": str, "description": "The prompt or prompts to avoid during image generation. Ignored when guidance_scale is less than 1."},
            "height": {"type": int, "default": 480, "description": "The height in pixels of the generated image."},
            "width": {"type": int, "default": 832, "description": "The width in pixels of the generated image."},
            "num_frames": {"type": int, "default": 81, "description": "The number of frames in the generated video."},
            "num_inference_steps": {"type": int, "default": 50, "description": "The number of denoising steps. More steps usually lead to higher quality at the expense of slower inference."},
            "guidance_scale": {"type": float, "default": 5.0, "description": "Guidance scale for classifier-free diffusion. Higher values encourage generation to be closely linked to the text prompt."},
            "num_videos_per_prompt": {"type": int, "default": 1, "description": "The number of videos to generate per prompt."},
            "generator": {"type": "torch.Generator", "description": "A torch.Generator or List[torch.Generator] to make generation deterministic."},
            "latents": {"type": "torch.Tensor", "description": "Pre-generated noisy latents to be used as inputs for image generation."},
            "prompt_embeds": {"type": "torch.Tensor", "description": "Pre-generated text embeddings, used as an alternative to the 'prompt' argument."},
            "output_type": {"type": str, "default": "np", "description": "The output format of the generated image. Choose between 'PIL.Image' or 'np.array'."},
            "return_dict": {"type": bool, "default": True, "description": "Whether or not to return a WanPipelineOutput object instead of a plain tuple."},
            "attention_kwargs": {"type": dict, "description": "A kwargs dictionary passed to the AttentionProcessor."},
            "callback_on_step_end": {"type": "Callable", "description": "A function called at the end of each denoising step during inference."},
            "callback_on_step_end_tensor_inputs": {"type": list, "description": "The list of tensor inputs for the callback_on_step_end function."},
            "max_sequence_length": {"type": int, "default": 512, "description": "The maximum sequence length of the text encoder."}
        }
    },
    "wan2.1-vace-14b": {
        "required": {
            "prompt": {"type": str, "description": "Text prompt to guide generation"},
        },
        "optional": {
            "negative_prompt": {"type": str, "default": "", "description": "The prompt or prompts not to guide the image generation. Ignored when guidance_scale is less than 1."},
            "conditioning_scale": {"type": float, "default": 1.0, "description": "The scale applied to the control conditioning latent stream. Can be a float, List[float], or torch.Tensor."},
            "height": {"type": int, "default": 480, "description": "The height in pixels of the generated video."},
            "width": {"type": int, "default": 832, "description": "The width in pixels of the generated video."},
            "num_frames": {"type": int, "default": 81, "description": "The number of frames in the generated video."},
            "num_inference_steps": {"type": int, "default": 50, "description": "The number of denoising steps. More steps usually lead to higher quality at the expense of slower inference."},
            "guidance_scale": {"type": float, "default": 5.0, "description": "Guidance scale for classifier-free diffusion. Higher values encourage generation to be closely linked to the text prompt."},
            "num_videos_per_prompt": {"type": int, "default": 1, "description": "The number of videos to generate per prompt."},
            "generator": {"type": "torch.Generator", "description": "A torch.Generator or List[torch.Generator] to make generation deterministic."},
            "latents": {"type": "torch.Tensor", "description": "Pre-generated noisy latents to be used as inputs for generation."},
            "prompt_embeds": {"type": "torch.Tensor", "description": "Pre-generated text embeddings. Can be used to easily tweak text inputs."},
            "output_type": {"type": str, "default": "np", "description": "The output format of the generated image. Choose between 'PIL.Image' or 'np.array'."},
            "return_dict": {"type": bool, "default": True, "description": "Whether or not to return a WanPipelineOutput instead of a plain tuple."},
            "attention_kwargs": {"type": dict, "description": "A kwargs dictionary passed along to the AttentionProcessor."},
            "callback_on_step_end": {"type": "Callable", "description": "A function called at the end of each denoising step during inference."},
            "callback_on_step_end_tensor_inputs": {"type": list, "description": "The list of tensor inputs for the callback_on_step_end function."},
            "max_sequence_length": {"type": int, "default": 512, "description": "The maximum sequence length of the text encoder."},
            "flow_shift": {"type": float, "default": 3.0, "description": "a value that estimates motion between two frames. A larger flow shift focuses on high motion or transformation. A smaller flow shift focuses on stability."}
        }
    },
    "cogvideox-5b": {
        "required": {
            "prompt": {"type": str, "description": "Text prompt to guide generation"},
        },
        "optional": {
            "negative_prompt": {"type": str, "description": "The prompt or prompts not to guide video generation. Ignored if guidance_scale is less than 1."},
            "height": {"type": int, "default": 480, "description": "The height in pixels of the generated video."},
            "width": {"type": int, "default": 720, "description": "The width in pixels of the generated video."},
            "num_frames": {"type": int, "default": 48, "description": "Number of frames to generate."},
            "num_inference_steps": {"type": int, "default": 50, "description": "The number of denoising steps. More steps can improve quality but are slower."},
            "timesteps": {"type": list, "description": "Custom timesteps to use for the denoising process, must be in descending order."},
            "guidance_scale": {"type": float, "default": 7.0, "description": "Classifier-Free Diffusion guidance scale. Higher values align the video more closely with the prompt."},
            "num_videos_per_prompt": {"type": int, "default": 1, "description": "The number of videos to generate for each prompt."},
            "generator": {"type": "torch.Generator", "description": "A torch.Generator or List[torch.Generator] to make generation deterministic."},
            "latents": {"type": "torch.FloatTensor", "description": "Pre-generated noisy latents to be used as inputs for generation."},
            "prompt_embeds": {"type": "torch.FloatTensor", "description": "Pre-generated text embeddings, used as an alternative to the 'prompt' argument."},
            "negative_prompt_embeds": {"type": "torch.FloatTensor", "description": "Pre-generated negative text embeddings, used as an alternative to the 'negative_prompt' argument."},
            "output_type": {"type": str, "default": "pil", "description": "The output format of the generated video. Choose between 'pil' or 'np.array'."},
            "return_dict": {"type": bool, "default": True, "description": "Whether to return a StableDiffusionXLPipelineOutput object instead of a plain tuple."},
            "attention_kwargs": {"type": dict, "description": "A kwargs dictionary passed to the AttentionProcessor."},
            "callback_on_step_end": {"type": "Callable", "description": "A function called at the end of each denoising step during inference."},
            "callback_on_step_end_tensor_inputs": {"type": list, "description": "The list of tensor inputs for the callback_on_step_end function."},
            "max_sequence_length": {"type": int, "default": 226, "description": "Maximum sequence length in the encoded prompt."}
        }
    },
    "pusa-v1": {
        "required": {
            "prompt": {"type": str, "description": "Text prompt to guide generation"},
        },
        "optional": {
            "negative_prompt": {"type": str, "description": "The prompt or prompts not to guide video generation. Ignored if guidance_scale is less than 1."},
        }
    },
    "wan2.1-vace-14b-phantom-fusionx": {
        "required": {
            "prompt": {"type": str, "description": "Text prompt to guide generation"},
        },
        "optional": {
            "negative_prompt": {"type": str, "default": "", "description": "The prompt or prompts not to guide the image generation. Ignored when guidance_scale is less than 1."},
            "video": {"type": list, "description": "The input video (List[PIL.Image.Image]) to be used as a starting point for the generation."},
            "mask": {"type": list, "description": "The input mask (List[PIL.Image.Image]) that defines which video regions to condition on (black) and which to generate (white)."},
            "reference_images": {"type": list, "description": "A list of one or more reference images (List[PIL.Image.Image]) as extra conditioning for the generation."},
            "conditioning_scale": {"type": float, "default": 1.0, "description": "The scale applied to the control conditioning latent stream. Can be a float, List[float], or torch.Tensor."},
            "height": {"type": int, "default": 480, "description": "The height in pixels of the generated video."},
            "width": {"type": int, "default": 832, "description": "The width in pixels of the generated video."},
            "num_frames": {"type": int, "default": 81, "description": "The number of frames in the generated video."},
            "num_inference_steps": {"type": int, "default": 50, "description": "The number of denoising steps. More steps usually lead to higher quality at the expense of slower inference."},
            "guidance_scale": {"type": float, "default": 5.0, "description": "Guidance scale for classifier-free diffusion. Higher values encourage generation to be closely linked to the text prompt."},
            "num_videos_per_prompt": {"type": int, "default": 1, "description": "The number of videos to generate per prompt."},
            "generator": {"type": "torch.Generator", "description": "A torch.Generator or List[torch.Generator] to make generation deterministic."},
            "latents": {"type": "torch.Tensor", "description": "Pre-generated noisy latents to be used as inputs for generation."},
            "prompt_embeds": {"type": "torch.Tensor", "description": "Pre-generated text embeddings. Can be used to easily tweak text inputs."},
            "output_type": {"type": str, "default": "np", "description": "The output format of the generated image. Choose between 'PIL.Image' or 'np.array'."},
            "return_dict": {"type": bool, "default": True, "description": "Whether or not to return a WanPipelineOutput instead of a plain tuple."},
            "attention_kwargs": {"type": dict, "description": "A kwargs dictionary passed along to the AttentionProcessor."},
            "callback_on_step_end": {"type": "Callable", "description": "A function called at the end of each denoising step during inference."},
            "callback_on_step_end_tensor_inputs": {"type": list, "description": "The list of tensor inputs for the callback_on_step_end function."},
            "max_sequence_length": {"type": int, "default": 512, "description": "The maximum sequence length of the text encoder."},
            "flow_shift": {"type": float, "default": 3.0, "description": "a value that estimates motion between two frames. A larger flow shift focuses on high motion or transformation. A smaller flow shift focuses on stability."}
        }
    },
    # Add more models as needed
}



"""TioMagic - A unified interface for video generation across multiple providers.

This module provides a high-level API for various video generation tasks including
text-to-video, image-to-video, interpolation, and pose guidance. It supports multiple
backend providers (local, modal, baseten) and handles job management, validation,
and status tracking.

Example:
    Basic usage::

        from tiomagic import tm
        
        # Configure provider
        tm.configure(provider="modal", api_key="your-api-key")
        
        # Generate video from text
        job = tm.text_to_video(
            model="cogvideox",
            required_args={"prompt": "A cat playing piano"}
        )
        
        # Check status
        tm.check_generation_status(job.job_id)

Attributes:
    tm: Pre-initialized TioMagic instance for convenience.
"""
from typing import Any, Dict
from uuid import uuid4

from .core.schemas import FEATURE_SCHEMAS
from .core.registry import registry
from .core.config import Configuration
from .core._jobs import Job, JobStatus
from .core._validation import validate_parameters
from .core._utils import create_timestamp
from .core.errors import (
    UnknownModelError, UnknownProviderError, ValidationError,
    JobExecutionError, ResourceNotFoundError,
)
from dotenv import load_dotenv
load_dotenv()

class TioMagic:
    """Main interface for video generation operations.
    
    This class provides methods for various video generation tasks and manages
    provider configuration, job tracking, and error handling. It supports multiple
    backend providers and models through a registry system.
    
    Attributes:
        _config: Configuration instance managing provider settings and API keys.
    """
    def __init__(self):
        """Initialize TioMagic with default configuration."""
        self._config = Configuration()

    def configure(self, provider=None, api_key=None, model_path=None, gpu=None, timeout=None, scaledown_window=None):
        """Configure the video generation provider and credentials.
        
        Args:
            provider (str, optional): Provider name ('local', 'modal', 'baseten').
            api_key (str, optional): API key for the specified provider.
            model_path (str, optional): Path to local model files (for local provider).
            gpu (GPUType, optional): GPU type to run on modal.
            timeout (int, optional): Measure of execution time on modal.
            scaledown_window (int, optional): How long container is warm for after completion of a task modal. Inputs will execute more quickly on a warm container compared to a cold one
            flow_shift (float, optional): Set flow shift for models that accept flow shift

        Example:
            >>> tm.configure(provider="modal", api_key="sk-...")
            >>> tm.configure(provider="local", model_path="/path/to/model")
        """
        if provider:
            self._config.set_provider(provider)
        if api_key:
            self._config.set_api_key(provider, api_key)
        if model_path:
            self._config.set_model_path(provider, model_path)
        if provider == 'modal' and (gpu or timeout or scaledown_window):
            self._config.set_modal_options(gpu, timeout, scaledown_window)

    def text_to_video(self, model=None, required_args: Dict[str, Any]= None, **kwargs):
        """Generate video from text prompt.
        
        Creates a video generation job that converts text descriptions into video content
        using the specified model and provider.
        
        Args:
            model (str, optional): Model identifier (e.g., 'cogvideox', 'stable-video').
                If None, uses provider's default model.
            required_args (Dict[str, Any], optional): Required parameters including:
                - prompt (str): Text description of the desired video.
                Additional model-specific parameters may be required.
            **kwargs: Additional optional parameters specific to the model/provider.
            
        Returns:
            Job: Job object for tracking generation progress and retrieving results.
        """
        provider = self._config.get_provider()
        impl_class = registry.get_implementation("text_to_video", model, provider)
        implementation = self._create_implementation(impl_class, provider)
        print("--> Implementation: ", implementation)

        valid, params, error_msg = validate_parameters("text_to_video", model, required_args, kwargs)
        if not valid:
            # Parse error message to provide more specific validation error
            field = "unknown"
            if "prompt" in error_msg:
                field = "prompt"
            elif "model" in error_msg:
                field = "model"
            
            raise ValidationError(
                field=field,
                message=error_msg,
                value=required_args.get(field) if required_args else None
            )
        
        # if modal provider, add modal_options into kwargs
        print("PROVIDER", provider)
        if provider == 'modal':
            print("PROVIDER MODAL, SET OPTIONS")

            kwargs['modal_options'] = self._config.get_modal_options()

        print("--> Validated parameters: ", params)

        # Start job and return job object for tracking
        job_id = str(uuid4())
        job = Job(
            job_id=job_id, 
            feature="text_to_video", 
            model=model, 
            provider=provider,
        )
        job.save()
        try:

            print(f"--> Text to video create new job with provider {provider} and model {model}")
            job.start(lambda: implementation.generate(required_args, **kwargs))
            print(f"--> Generation started! Job ID: {job_id}")
        except Exception as e:
            job.update(status=JobStatus.FAILED)
            raise JobExecutionError(
                job_id=job_id,
                reason=f"Failed to start text-to-video generation: {str(e)}",
                provider_error=str(e)
            )

        return job

    def image_to_video(self, model=None, required_args: Dict[str, Any]= None, **kwargs):
        """Generate video from an input image.
        
        Creates a video generation job that animates or extends a static image into
        video content using the specified model and provider.
        
        Args:
            model (str, optional): Model identifier for image-to-video generation.
                If None, uses provider's default model.
            required_args (Dict[str, Any], optional): Required parameters including:
                - image (str): Path or URL to the input image.
                - prompt (str, optional): Text description to guide video generation.
                Additional model-specific parameters may be required.
            **kwargs: Additional optional parameters specific to the model/provider.
            
        Returns:
            Job: Job object for tracking generation progress and retrieving results.
        """
        provider = self._config.get_provider()
        impl_class = registry.get_implementation("image_to_video", model, provider)
        implementation = self._create_implementation(impl_class, provider)
        print("--> Implementation: ", implementation)

        valid, params, error_msg = validate_parameters("image_to_video", model, required_args, kwargs)
        if not valid:
            # Parse error message to provide more specific validation error
            field = "unknown"
            if "prompt" in error_msg:
                field = "prompt"
            elif "model" in error_msg:
                field = "model"
            
            raise ValidationError(
                field=field,
                message=error_msg,
                value=required_args.get(field) if required_args else None
            )
        # if modal provider, add modal_options into kwargs
        if provider == 'modal':
            kwargs['modal_options'] = self._config.get_modal_options()

        print("--> Validated parameters: ", params)
        

        job_id = str(uuid4())
        job = Job(
            job_id=job_id,
            feature="image_to_video",
            model=model,
            provider=provider
        )
        job.save()
        try:
            print(f"--> Image to video create new job with provider {provider} and model {model}")
            job.start(lambda: implementation.generate(required_args, **kwargs))
            print(f"--> Generation started! Job ID: {job_id}")
        except Exception as e:
            job.update(status=JobStatus.FAILED)
            raise JobExecutionError(
                job_id=job_id,
                reason=f"Failed to start image-to-video generation: {str(e)}",
                provider_error=str(e)
            )

        return job

    def interpolate(self, model=None, required_args: Dict[str, Any] = None, **kwargs):
        """Generate interpolated video between keyframes or images.
        
        Creates a video generation job that interpolates between multiple input frames
        to create smooth transitions or in-between frames.
        
        Args:
            model (str, optional): Model identifier for interpolation.
                If None, uses provider's default interpolation model.
            required_args (Dict[str, Any], optional): Required parameters including:
                - frames (List[str]): List of paths/URLs to keyframe images.
                - prompt (str, optional): Text description to guide interpolation.
                Additional model-specific parameters may be required.
            **kwargs: Additional optional parameters specific to the model/provider.
            
        Returns:
            Job: Job object for tracking generation progress and retrieving results.
        """
        provider = self._config.get_provider()
        impl_class = registry.get_implementation("interpolate", model, provider)
        implementation = self._create_implementation(impl_class, provider)
        print("--> Implementation: ", implementation)

        valid, params, error_msg = validate_parameters("interpolate", model, required_args, kwargs)
        if not valid:
            # Parse error message to provide more specific validation error
            field = "unknown"
            if "prompt" in error_msg:
                field = "prompt"
            elif "model" in error_msg:
                field = "model"
            
            raise ValidationError(
                field=field,
                message=error_msg,
                value=required_args.get(field) if required_args else None
            )
        # if modal provider, add modal_options into kwargs
        if provider == 'modal':
            kwargs['modal_options'] = self._config.get_modal_options()

        print("--> Validated parameters: ", params)

        job_id = str(uuid4())
        job = Job(
            job_id=job_id,
            feature="interpolate",
            model=model,
            provider=provider
        )
        job.save()

        try:
            print(f"--> Interpolate create new job with provider {provider} and model {model}")
            job.start(lambda: implementation.generate(
                required_args,
                **kwargs))
            print(f"--> Generation started! Job ID: {job_id}")
        except Exception as e:
            job.update(status=JobStatus.FAILED)
            raise JobExecutionError(
                job_id=job_id,
                reason=f"Failed to start interpolation generation: {str(e)}",
                provider_error=str(e)
            )

        return job

    def pose_guidance(self, model=None, required_args: Dict[str, Any] = None, **kwargs):
        """Generate video with pose-guided animation.
        
        Creates a video generation job that uses pose information (skeleton, keypoints)
        to guide character or object animation in the generated video.
        
        Args:
            model (str, optional): Model identifier for pose-guided generation.
                If None, uses provider's default pose guidance model.
            required_args (Dict[str, Any], optional): Required parameters including:
                - pose_data: Pose information (format depends on model).
                - prompt (str): Text description of the desired animation.
                - reference_image (str, optional): Reference image for appearance.
                Additional model-specific parameters may be required.
            **kwargs: Additional optional parameters specific to the model/provider.
            
        Returns:
            Job: Job object for tracking generation progress and retrieving results.
        """
        provider = self._config.get_provider()
        impl_class = registry.get_implementation("pose_guidance", model, provider)
        implementation = self._create_implementation(impl_class, provider)
        print("--> Implementation: ", implementation)

        valid, params, error_msg = validate_parameters("pose_guidance", model, required_args, kwargs)
        if not valid:
           # Parse error message to provide more specific validation error
            field = "unknown"
            if "prompt" in error_msg:
                field = "prompt"
            elif "model" in error_msg:
                field = "model"
            
            raise ValidationError(
                field=field,
                message=error_msg,
                value=required_args.get(field) if required_args else None
            )
        # if modal provider, add modal_options into kwargs
        if provider == 'modal':
            kwargs['modal_options'] = self._config.get_modal_options()

        print("--> Validated parameters: ", params)

        job_id = str(uuid4())
        job = Job(
            job_id=job_id,
            feature="pose_guidance",
            model=model,
            provider=provider
        )
        job.save()

        try:
            print(f"--> Pose guidance create new job with provider {provider} and model {model}")
            job.start(lambda: implementation.generate(
                required_args,
                **kwargs))
            print(f"--> Generation started! Job ID: {job_id}")
        except Exception as e:
            job.update(status=JobStatus.FAILED)
            raise JobExecutionError(
                job_id=job_id,
                reason=f"Failed to start pose guidance generation: {str(e)}",
                provider_error=str(e)
            )

        return job
    def check_generation_status(self, job_id, returnJob = False):
        """Check the current status of a video generation job.
        
        Queries the provider for the current status of a running job and updates
        the local job record. This method can be called periodically to monitor
        long-running generation tasks.
        
        Args:
            job_id (str): Unique identifier of the job to check.
            
        Returns:
            None: Updates are made to the job record in place.
        """
        job = Job.get_job(job_id)
        if not job:
            raise ResourceNotFoundError(
                resource_type="Job",
                resource_id=job_id,
                location="job storage"
            )

        # Check current status
        print("--> Check job: ", job.job_id)

        if job.generation and job.generation["call_id"]:
            try:
                # assumption it is modal for now
                impl_class = registry.get_implementation(job.feature, job.model, job.provider)
                implementation = self._create_implementation(impl_class, job.provider)
                job.check_status(lambda: implementation.check_generation_status(job.generation))
                job.update(last_updated= create_timestamp())
                # checks generation status and updates generation_log.json
                if returnJob:
                    return job
                else:
                    return

            except (UnknownModelError, UnknownProviderError):
                raise
            except Exception as e:
                job.update(status=JobStatus.FAILED)
                job.save()
                raise JobExecutionError(
                    job_id=job_id,
                    reason=f"Failed to check generation status: {str(e)}",
                    provider_error=str(e)
                )
        else:
            # Job exists but has no generation info
            raise JobExecutionError(
                job_id=job_id,
                reason="Job exists but has no generation information to check"
            )
    
    def cancel_job(self, job_id):
        """Cancel an active video generation job.
        
        Attempts to cancel a running job with the provider. The success of cancellation
        depends on the provider's capabilities and the current state of the job.
        
        Args:
            job_id (str): Unique identifier of the job to cancel.
            
        Returns:
            None: Job status is updated to reflect cancellation.
        """
        job = Job.get_job(job_id)
        if not job:
            raise ResourceNotFoundError(
                resource_type="Job",
                resource_id=job_id,
                location="job storage"
            )
        print("--> Cancel job: ", job.job_id)
        if job.generation and job.generation["call_id"]:
            try:
                impl_class = registry.get_implementation(job.feature, job.model, job.provider)
                implementation = self._create_implementation(impl_class, job.provider)
                job.cancel_job(lambda: implementation.cancel_job(job.generation))
                job.update(last_updated= create_timestamp())
            except (UnknownModelError, UnknownProviderError):
                raise
            except Exception as e:
                job.update(status=JobStatus.FAILED)
                job.save()
                raise JobExecutionError(
                    job_id=job_id,
                    reason=f"Failed to cancel job: {str(e)}",
                    provider_error=str(e)
                )
        else:
            raise JobExecutionError(
                job_id=job_id,
                reason="Job has no active generation to cancel"
            )
        
    def get_providers(self):
        """List all providers available."""
        return registry.get_providers()
    def get_models(self, feature, provider):
        """List all models available for a provider and feature"""
        return registry.get_models(feature=feature, provider=provider)
    def get_schema(self, feature, model):
        """List schema for particular implementation"""
        schema = FEATURE_SCHEMAS[feature][model]
        print(f"schema for {feature} {model}: {schema}")
    def list_implementations(self):
        """List all implementations available
        Models listed and sorted by provider or feature
        """
        providers = registry.get_providers()
        features = registry.get_features()
        for provider in providers:
            print(f"{provider}: ")
            for feature in features:
                feature_models = registry.get_models(feature=feature, provider=provider)
                if len(feature_models) > 0:
                    print(f"\t{feature}: ")
                    for model in feature_models:
                        print(f"\t\t{model}")

    
    def _create_implementation(self, impl_class, provider):
        """Create an implementation instance with appropriate config.
        
        Internal method to instantiate provider-specific implementations
        with the correct configuration parameters.
        """
        if provider == "local":
            # return impl_class(model_path=self._config.get_model_path())
            return impl_class()
        elif provider == "modal":
            return impl_class(api_key=self._config.get_api_key("modal"))
        elif provider == "baseten":
            return impl_class(api_key=self._config.get_api_key("baseten"))
        else:
            return impl_class()

tm = TioMagic()
"""Pre-initialized TioMagic instance for direct module usage.

This allows users to import and use TioMagic without explicit instantiation:

Example:
    >>> from tiomagic import tm
    >>> tm.configure(provider="modal", api_key="...")
    >>> job = tm.text_to_video(...)
"""


import modal
from pathlib import Path
from fastapi.responses import JSONResponse

from .base import GPUType, GenericWebAPI, ModalProviderBase
from typing import Any, Dict
from ...core.registry import registry
from ...core._utils import load_image_robust, is_local_path, local_image_to_base64, create_timestamp, extract_image_dimensions
from ...core.constants import FeatureType
from ...core.schemas import FEATURE_SCHEMAS
from ...core.errors import (
    DeploymentError, ProcessingError
)


APP_NAME = "wan-2.1-flf2v-14b-720p"
MODEL_NAME = "wan2.1-flf2v-14b-720p"
CACHE_NAME = f"{APP_NAME}-cache"
CACHE_PATH = "/cache"
OUTPUTS_NAME = f"{APP_NAME}-outputs"
OUTPUTS_PATH = "/outputs"

MODEL_ID = "Wan-AI/Wan2.1-FLF2V-14B-720P-diffusers"

GPU_CONFIG: GPUType = GPUType.A100_80GB
TIMEOUT: int = 1800 # 30 minutes
SCALEDOWN_WINDOW: int = 900 # stay idle for 15 minutes before scaling down

image = (
    modal.Image.debian_slim(python_version="3.11")
    .apt_install("git")
    .pip_install(
        "git+https://github.com/huggingface/diffusers.git",
        "torch>=2.4.0",
        "torchvision>=0.19.0",
        "opencv-python>=4.9.0.80",
        "diffusers>=0.31.0",
        "transformers>=4.49.0",
        "tokenizers>=0.20.3",
        "accelerate>=1.1.1",
        "tqdm",
        "imageio",
        "easydict",
        "ftfy",
        "dashscope",
        "imageio-ffmpeg",
        "gradio>=5.0.0",
        "numpy>=1.23.5,<2",
        "fastapi",
    ).env({"HF_HUB_CACHE": CACHE_PATH})
)

cache_volume = modal.Volume.from_name(CACHE_NAME, create_if_missing=True)
outputs_volume = modal.Volume.from_name(OUTPUTS_NAME, create_if_missing=True)

app = modal.App(APP_NAME)


@app.cls(
    image=image,
    gpu=GPU_CONFIG,
    secrets=[modal.Secret.from_name("huggingface-secret")],
    volumes={CACHE_PATH: cache_volume, OUTPUTS_PATH: outputs_volume},
    timeout=TIMEOUT,
    scaledown_window=SCALEDOWN_WINDOW,
)
class Interpolate:
    @modal.enter()
    def load_models(self):
        import torch
        from diffusers import AutoencoderKLWan, WanImageToVideoPipeline
        from transformers import CLIPVisionModel
        import time

        print("Loading models...")
        start_time = time.time()
        print(f"✅ {time.time() - start_time:.2f}s: Starting model load...")

        try:
            print(f"✅ {time.time() - start_time:.2f}s: Loading image_encoder...")
            image_encoder = CLIPVisionModel.from_pretrained(MODEL_ID, subfolder="image_encoder", torch_dtype=torch.float32)
            print(f"✅ {time.time() - start_time:.2f}s: image_encoder loaded.")

            print(f"✅ {time.time() - start_time:.2f}s: Loading VAE...")
            vae = AutoencoderKLWan.from_pretrained(MODEL_ID, subfolder="vae", torch_dtype=torch.float32)
            print(f"✅ {time.time() - start_time:.2f}s: VAE loaded.")

            print(f"✅ {time.time() - start_time:.2f}s: Creating pipeline...")

            self.pipe = WanImageToVideoPipeline.from_pretrained(
                MODEL_ID, 
                vae=vae, 
                image_encoder=image_encoder, 
                torch_dtype=torch.bfloat16
            )
            self.pipe.to("cuda")
            print(f"✅ {time.time() - start_time:.2f}s: Pipeline ready. Models loaded successfully.")
        except Exception as e:
            print(f"❌ {time.time() - start_time:.2f}s: An error occurred: {e}")
            raise
        print("Models loaded successfully.")
    def aspect_ratio_resize(self, image, max_area=720 * 1280):
        import numpy as np
        aspect_ratio = image.height / image.width
        mod_value = self.pipe.vae_scale_factor_spatial * self.pipe.transformer.config.patch_size[1]
        height = round(np.sqrt(max_area * aspect_ratio)) // mod_value * mod_value
        width = round(np.sqrt(max_area / aspect_ratio)) // mod_value * mod_value
        image = image.resize((width, height))
        return image, height, width

    def center_crop_resize(self, image, height, width):
        import torchvision.transforms.functional as TF

        # Calculate resize ratio to match first frame dimensions
        resize_ratio = max(width / image.width, height / image.height)

        # Resize the image
        width = round(image.width * resize_ratio)
        height = round(image.height * resize_ratio)
        size = [width, height]
        image = TF.center_crop(image, size)

        return image, height, width
    @modal.method()
    def generate(self, data: Dict[str, Any]):
        from diffusers.utils import export_to_video
        interpolate_schema = FEATURE_SCHEMAS["interpolate"][MODEL_NAME]

        first_frame = data.get('first_frame')
        last_frame = data.get('last_frame')
        # LOAD_IMAGE these frames
        height = data.get('height', interpolate_schema["optional"]["height"]["default"])
        width = data.get('width', interpolate_schema["optional"]["height"]["default"])

        data['first_frame'], height, width = self.aspect_ratio_resize(first_frame)
        if last_frame.size != data['first_frame'].size:
            data['last_frame'], _, _ = self.center_crop_resize(last_frame, height, width)
        data['image'] = data.pop('first_frame')
        data['last_image'] = data.pop('last_frame')
        output = self.pipe(**data).frames[0]

        timestamp = create_timestamp()
        mp4_name = f"{MODEL_NAME}-interpolate-output_{timestamp}.mp4"
        mp4_path = Path(OUTPUTS_PATH) / mp4_name
        export_to_video(output, str(mp4_path), fps=16)
        outputs_volume.commit()

        with open(mp4_path, "rb") as f:
            video_bytes = f.read()

        return video_bytes
    @staticmethod
    def handle_web_inference(data: dict):
        first_frame = data.get("first_frame")
        last_frame = data.get("last_frame")

        try:
            # load_image_robust can handle both URLs and base64 strings
            first_frame = load_image_robust(first_frame)
            last_frame = load_image_robust(last_frame)   
            data['first_frame'] = first_frame
            data['last_frame'] = last_frame
            if 'height' not in data and 'width' not in data:
                data = extract_image_dimensions(first_frame, data)
        except Exception as e:
            raise ProcessingError(
                media_type="image",
                operation="load and process",
                reason=str(e),
                file_path=data.get("image") if isinstance(data.get("image"), str) else None
            )

        # Create Interpolate instance and call generate
        try:
            interpolate_instance = Interpolate()
            call = interpolate_instance.generate.spawn(data)
        except Exception as e:
            raise DeploymentError(
                service="Modal",
                reason=f"Failed to spawn Interpolate job: {str(e)}",
                app_name=APP_NAME
            )

        return JSONResponse({"call_id": call.object_id, "feature_type": FeatureType.INTERPOLATE})

class Wan21FlfvInterpolate14b720p(ModalProviderBase):
    def __init__(self, api_key=None):
        super().__init__(api_key)
        self.app_name = APP_NAME
        self.modal_app = app
        self.modal_class_name = "Interpolate"
    def _prepare_payload(self, required_args: Dict[str, Any], **kwargs) -> Dict[str, Any]:
        """Prepare payload specific to Wan2.1 Vace Interpolate model.
        Break out required args into payload
        """
        payload = super()._prepare_payload(required_args, **kwargs)
        # payload = {"prompt": required_args['prompt']}
        payload["feature_type"] = FeatureType.INTERPOLATE

        if payload['first_frame'] is None or payload['last_frame'] is None:
            raise ValueError("Arguments 'first_frame' and 'last_frame' are required for Interpolation Video generation")

        if is_local_path(payload['first_frame']):
            # Convert local image to base64
            payload["first_frame"] = local_image_to_base64(payload['first_frame'])
        if is_local_path(payload['last_frame']):
            # Convert local image to base64
            payload["last_frame"] = local_image_to_base64(payload['last_frame'])

        return payload

# Create a subclass with the handlers
class WebAPI(GenericWebAPI):
    feature_handlers = {
        FeatureType.INTERPOLATE: Interpolate
    }

# Apply Modal decorator
WebAPI = app.cls(
    image=image,
    gpu=GPU_CONFIG,
    secrets=[modal.Secret.from_name("huggingface-secret")],
    volumes={CACHE_PATH: cache_volume, OUTPUTS_PATH: outputs_volume},
    timeout=TIMEOUT,
    scaledown_window=SCALEDOWN_WINDOW,
)(WebAPI)


registry.register(
    feature=FeatureType.INTERPOLATE,
    model="wan2.1-flf2v-14b-720p",
    provider="modal",
    implementation=Wan21FlfvInterpolate14b720p
)



import time
import os
from typing import Any, Dict

import requests

from ...core.errors import GenerationError, ProcessingError
from ...core._utils._utils import create_timestamp

from ...core.registry import registry
from ...core._utils import is_local_path
from ...core.constants import Generation
from .base import LocalProviderBase
from ...core.constants import FeatureType

from lumaai import LumaAI
# async options available, check Luma Documentation

APP_NAME = "luma-ray-2"
MODEL_NAME = "ray-2"

class LumaRay2I2V(LocalProviderBase):
    def __init__(self):
        super().__init__()
        api_key = os.getenv("LUMAAI_API_KEY")
        if not api_key:
            raise ValueError("LUMAAI_API_KEY not found in environment variables. Please set it in your .env file.")
        self.client = LumaAI(
            auth_token=os.environ.get("LUMAAI_API_KEY"),
        )
        self.app_name = APP_NAME
        self.feature = FeatureType.IMAGE_TO_VIDEO
        self.operations = None
    def _prepare_payload(self, required_args: Dict[str, Any], **kwargs) -> Dict[str, Any]:
        payload = super()._prepare_payload(required_args, **kwargs)
        payload["feature_type"] = FeatureType.IMAGE_TO_VIDEO

        if payload['image'] is None:
            raise ValueError("Argument 'image' is required for Image to Video generation")

        if is_local_path(payload['image']):
            raise ProcessingError(
                media_type="image",
                operation="load local image",
                reason="Luma AI API only takes image URLs, not local files"
            )
        
        keyframes = {
                "frame0": {
                    "type": "image",
                    "url": payload['image']
                }
            }
        payload['keyframes'] = keyframes

        return payload
    def generate(self, required_args: Dict[str, Any], **kwargs) -> Dict[str, Any]:
        self._validate_config()
        print(f"Running {self.app_name} generate with prompt: {required_args['prompt']}")

        generation_obj = Generation(
            timestamp=create_timestamp(),
            required_args=required_args,
            optional_args=kwargs
        )
        print('-->generation object done', generation_obj)

        try:
            payload = self._prepare_payload(required_args, **kwargs)
            print('-->payload done')

            generation = self.client.generations.create(
                model=MODEL_NAME,
                prompt = payload['prompt'],
                keyframes = payload['keyframes']
            )
            print("--> generation creation done", generation)

            completed = False
            while not completed:
                generation = self.client.generations.get(id=generation.id)
                if generation.state == "completed":
                    completed = True
                elif generation.state == "failed":
                    generation_obj.update(
                        status="failed",
                        message=f"Generation failed: {generation.failure_reason}"
                    )
                    raise GenerationError(
                        app_name=APP_NAME,
                        model=MODEL_NAME,
                        feature=FeatureType.IMAGE_TO_VIDEO,
                        reason=f"{generation.failure_reason}",
                    )
                time.sleep(3)
            save_video(generation, generation_obj)

        except Exception as e:
            print(f"Error in generate: {str(e)}")
            generation_obj.update(message=f"Error in generate: {str(e)}")

            raise GenerationError(app_name=APP_NAME, 
                                  model=MODEL_NAME,
                                  reason=str(e)
                                  )

class LumaRay2Interpolate(LocalProviderBase):
    def __init__(self):
        super().__init__()
        api_key = os.getenv("LUMAAI_API_KEY")
        if not api_key:
            raise ValueError("LUMAAI_API_KEY not found in environment variables. Please set it in your .env file.")
        self.client = LumaAI(
            auth_token=os.environ.get("LUMAAI_API_KEY"),
        )
        self.app_name = APP_NAME
        self.feature = FeatureType.INTERPOLATE
        self.operations = None
    def _prepare_payload(self, required_args: Dict[str, Any], **kwargs) -> Dict[str, Any]:
        payload = super()._prepare_payload(required_args, **kwargs)
        payload["feature_type"] = FeatureType.INTERPOLATE

        if payload['first_frame'] is None:
            raise ValueError("Argument 'first_frame' is required for Interpolate generation")
        if payload['last_frame'] is None:
            raise ValueError("Argument 'last_frame' is required for Interpolate generation")

        if is_local_path(payload['first_frame']) or is_local_path(payload['last_frame']):
            raise ProcessingError(
                media_type="image",
                operation="load local image",
                reason="Luma AI API only takes image URLs, not local files"
            )
        
        keyframes = {
                "frame0": {
                    "type": "image",
                    "url": payload['first_frame']
                },
                "frame1": {
                    "type": "image",
                    "url": payload['last_frame']
                }
            }
        payload['keyframes'] = keyframes

        return payload
    def generate(self, required_args: Dict[str, Any], **kwargs) -> Dict[str, Any]:
        self._validate_config()
        print(f"Running {self.app_name} generate with prompt: {required_args['prompt']}")

        generation_obj = Generation(
            timestamp = create_timestamp(),
            required_args=required_args,
            optional_args=kwargs
        )
        print('-->generation object done', generation_obj)

        try:
            payload = self._prepare_payload(required_args, **kwargs)
            print('-->payload done')

            generation = self.client.generations.create(
                model=MODEL_NAME,
                prompt = payload['prompt'],
                keyframes = payload['keyframes']
            )
            print("--> generation creation done", generation)

            completed = False
            while not completed:
                generation = self.client.generations.get(id=generation.id)
                if generation.state == "completed":
                    completed = True
                elif generation.state == "failed":
                    generation_obj.update(
                        status="failed",
                        message=f"Generation failed: {generation.failure_reason}"
                    )
                    raise GenerationError(
                        app_name=APP_NAME,
                        model=MODEL_NAME,
                        feature=FeatureType.IMAGE_TO_VIDEO,
                        reason=f"{generation.failure_reason}",
                    )
                time.sleep(3)
            save_video(generation, generation_obj)
            
        except Exception as e:
            print(f"Error in generate: {str(e)}")
            generation_obj.update(message=f"Error in generate: {str(e)}")

            raise GenerationError(app_name=APP_NAME, 
                                  model=MODEL_NAME,
                                  reason=str(e)
                                  )

def save_video(generation, generation_obj):
    video_url = generation.assets.video

    response = requests.get(video_url, stream=True)

    # Save video to file
    timestamp = create_timestamp()
    video_filename = f"{MODEL_NAME}_output_{timestamp}.mp4"

    # Get the directory of this file and navigate to repo root
    current_dir = os.path.dirname(os.path.abspath(__file__))
    repo_root = os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(current_dir))))
    output_videos_dir = os.path.join(repo_root, "output_videos")
    video_path = os.path.join(output_videos_dir, video_filename)

    # Create output_videos directory if it doesn't exist
    os.makedirs(output_videos_dir, exist_ok=True)
    
    # Save the video to the output_videos directory
    with open(video_path, 'wb') as file:
        file.write(response.content)
    print(f"File downloaded as {video_path}")

    generation_obj.update(
            call_id=generation.id,
            status="completed",
            message=f"Video generated and saved to {video_path}",
            result_video=video_path
        )

registry.register(
    feature=FeatureType.IMAGE_TO_VIDEO,
    model="luma-ray-2",
    provider="local",
    implementation=LumaRay2I2V
) 
registry.register(
    feature=FeatureType.INTERPOLATE,
    model="luma-ray-2",
    provider="local",
    implementation=LumaRay2Interpolate
)


import modal
from pathlib import Path
from fastapi.responses import JSONResponse

from ...core.errors import DeploymentError

from .base import GPUType, GenericWebAPI, ModalProviderBase
from typing import Any, Dict
from ...core.registry import registry
from ...core._utils import create_timestamp
from ...core.constants import FeatureType

# --- Configuration ---
APP_NAME = "cogvideox-5b"
CACHE_PATH = "/cache"
CACHE_NAME = f"{APP_NAME}-cache"
OUTPUTS_NAME = f"{APP_NAME}-outputs"
OUTPUTS_PATH = "/outputs"

COGVIDEOX_MODEL_ID = "THUDM/CogVideoX-5b"

GPU_CONFIG: GPUType = GPUType.A100_80GB
TIMEOUT: int = 1800 # 30 minutes
SCALEDOWN_WINDOW: int = 900 # 15 minutes

image = (
    modal.Image.debian_slim(python_version="3.11")
    .apt_install("git")
    .pip_install(
        "diffusers>=0.32.1",
        "accelerate>=1.1.1",
        "transformers>=4.46.2",
        "numpy==1.26.0",
        "torch>=2.5.0",
        "torchvision>=0.20.0",
        "sentencepiece>=0.2.0",
        "SwissArmyTransformer>=0.4.12",
        "gradio>=5.5.0",
        "imageio>=2.35.1",
        "imageio-ffmpeg>=0.5.1",
        "openai>=1.54.0",
        "moviepy>=2.0.0",
        "scikit-video>=1.1.11",
        "pydantic>=2.10.3",
        "torchao"
    ).env({"HF_HUB_CACHE": CACHE_PATH})
)

cache_volume = modal.Volume.from_name(CACHE_NAME, create_if_missing=True)
outputs_volume = modal.Volume.from_name(OUTPUTS_NAME, create_if_missing=True)
app = modal.App(APP_NAME)

@app.cls(
    image=image,
    gpu=GPU_CONFIG,
    secrets=[modal.Secret.from_name("huggingface-secret")],
    volumes={CACHE_PATH: cache_volume, OUTPUTS_PATH: outputs_volume},
    timeout=TIMEOUT,
    scaledown_window=SCALEDOWN_WINDOW,
)
class T2V:
    @modal.enter()
    def load_models(self):
        import torch
        from diffusers import CogVideoXPipeline, AutoModel
        from diffusers.quantizers import PipelineQuantizationConfig
        from diffusers.quantizers.quantization_config import TorchAoConfig

        print("Loading models...")

        pipeline_quant_config = PipelineQuantizationConfig(
            quant_mapping={
                "transformer": TorchAoConfig(quant_type="int8wo")
            }
        )

        # fp8 layerwise weight-casting
        transformer = AutoModel.from_pretrained(
            "THUDM/CogVideoX-5b",
            subfolder="transformer",
            torch_dtype=torch.bfloat16
        )
        transformer.enable_layerwise_casting(
            storage_dtype=torch.float8_e4m3fn, compute_dtype=torch.bfloat16
        )

        self.pipe = CogVideoXPipeline.from_pretrained(
            "THUDM/CogVideoX-5b",
            transformer=transformer,
            quantization_config=pipeline_quant_config,
            torch_dtype=torch.bfloat16
        )
        self.pipe.to("cuda")

        # model-offloading
        self.pipe.enable_model_cpu_offload()
        print("✅ Models loaded successfully.")
    @modal.method()
    def generate(self, data: Dict[str, Any]):
        from diffusers.utils import export_to_video
        frames = self.pipe(**data).frames[0]

        timestamp = create_timestamp()
        mp4_name = f"cogvideox-5b-output_{timestamp}.mp4"
        mp4_path = Path(OUTPUTS_PATH) / mp4_name
        export_to_video(frames, str(mp4_path), fps=16)
        outputs_volume.commit()

        with open(mp4_path, "rb") as f:
            video_bytes = f.read()

        return video_bytes
    @staticmethod
    def handle_web_inference(data: Dict[str, Any]):
        """handle text-to-video generation"""
        try:
            t2v_instance = T2V()
            call = t2v_instance.generate.spawn(data)
        except Exception as e:
            raise DeploymentError(
                service="Modal",
                reason=f"Failed to spawn T2V job: {str(e)}",
                app_name=APP_NAME
            )

        return JSONResponse({"call_id": call.object_id, "feature_type": FeatureType.TEXT_TO_VIDEO})

class CogVideoXTextToVideo5B(ModalProviderBase):
    def __init__(self, api_key=None):
        super().__init__(api_key)
        self.app_name = APP_NAME
        self.modal_app = app
        self.modal_class_name = "T2V"
    def _prepare_payload(self, required_args, **kwargs) -> Dict[str, Any]:
        """Prepare payload specific to Wan2.1 Vace model."""
        payload = super()._prepare_payload(required_args, **kwargs)

        # Add feature_type for routing
        payload["feature_type"] = FeatureType.TEXT_TO_VIDEO

        return payload

# Create a subclass with the handlers
class WebAPI(GenericWebAPI):
    feature_handlers = {
        FeatureType.TEXT_TO_VIDEO: T2V,
    }

# Apply Modal decorator
WebAPI = app.cls(
    image=image,
    gpu=GPU_CONFIG,
    secrets=[modal.Secret.from_name("huggingface-secret")],
    volumes={CACHE_PATH: cache_volume, OUTPUTS_PATH: outputs_volume},
    timeout=TIMEOUT,
    scaledown_window=SCALEDOWN_WINDOW,
)(WebAPI)

registry.register(
    feature=FeatureType.TEXT_TO_VIDEO,
    model="cogvideox-5b",
    provider="modal",
    implementation=CogVideoXTextToVideo5B,
)


from . import image_to_video, interpolate, pose_guidance, text_to_video

FEATURE_SCHEMAS = {
    "text_to_video": text_to_video.SCHEMAS,
    "image_to_video": image_to_video.SCHEMAS,
    "interpolate": interpolate.SCHEMAS,
    "pose_guidance": pose_guidance.SCHEMAS,
}


import time
from fastapi.responses import JSONResponse
import os
from typing import Any, Dict

from ...core.errors import GenerationError

from ...core.registry import registry
from ...core._utils import is_local_path, create_timestamp
from ...core.constants import Generation
from .base import LocalProviderBase
import base64
from ...core.constants import FeatureType


from google import genai
from google.genai import types

APP_NAME = "veo-2.0-generate-001"
MODEL_NAME = 'veo-2.0-generate-001'
# load_dotenv()
# GOOGLE_API_KEY = os.getenv("GOOGLE_API_KEY")
# if not GOOGLE_API_KEY:
#     raise ValueError("GOOGLE_API_KEY not found in environment variables")

class Veo20Generate001(LocalProviderBase):
    def __init__(self):
        super().__init__()
        api_key = os.getenv("GOOGLE_API_KEY")
        if not api_key:
            raise ValueError("GOOGLE_API_KEY not found in environment variables. Please set it in your .env file.")
        self.client = genai.Client(api_key=api_key)
        self.app_name = APP_NAME
        self.feature = FeatureType.IMAGE_TO_VIDEO
        self.operations = None


    def _prepare_payload(self, required_args: Dict[str, Any], **kwargs) -> Dict[str, Any]:
        """Prepare payload specific to Veo 2.0 Generate 001 model.
        Break out required args into payload
        """
        payload = super()._prepare_payload(required_args, **kwargs)
        payload["feature_type"] = FeatureType.IMAGE_TO_VIDEO

        if payload['image'] is None:
            raise ValueError("Argument 'image' is required for Image to Video generation")

        if is_local_path(payload['image']):
            print(f'Uploading local image: {payload["image"]}')
            # uploaded_file = self.client.files.upload(file=payload['image'])
            # print(f'Initial upload complete: {uploaded_file.name}')
            print("--> payload image path ", len(payload['image']))
            file = types.Image.from_file(location=payload['image'])
            payload['image'] = file


        #     # Wait a moment for processing
        #     import time
        #     time.sleep(2)

        #     # Retrieve the file to ensure it's fully processed
        #     retrieved_file = self.client.files.get(name=uploaded_file.name)

        #     # Check if file is active
        #     print(f'File state: {retrieved_file.state}')
        #     print(f'File details: name={retrieved_file.name}, mime_type={retrieved_file.mime_type}, size={retrieved_file.size_bytes}')

        #     # Wait for file to be active if needed
        #     max_attempts = 10
        #     attempt = 0
        #     while retrieved_file.state != "ACTIVE" and attempt < max_attempts:
        #         print(f"Waiting for file to become active... (attempt {attempt + 1})")
        #         time.sleep(2)
        #         retrieved_file = self.client.files.get(name=uploaded_file.name)
        #         attempt += 1

        #     if retrieved_file.state != "ACTIVE":
        #         raise ValueError(f"File upload failed or timed out. State: {retrieved_file.state}")

        #     payload['image'] = retrieved_file
        #     print(f'File ready for use: {retrieved_file.name}')

        # elif isinstance(payload['image'], str) and payload['image'].startswith('data:'):
        #     # Similar handling for base64...
        #     pass

        return payload
            # image = PIL.Image.open(payload['image'])
            # payload['image'] = image
            # base64_image = local_image_to_base64(payload['image'])
            # bytes_image = base64_to_bytes(base64_image)
            # payload['image'] = {"bytesBase64Encoded": bytes_image, "mimeType": "image/png"}

            # payload['image'] = base64_image.split(",", 1)[1]
            # payload["image"] = base64_to_bytes(base64_image)
        # return payload

    def generate(self, required_args: Dict[str, Any], **kwargs) -> Dict[str, Any]:
        """Generate video using Google Veo API
        """
        self._validate_config()

        from datetime import datetime
        print(f"Running {self.app_name} generate with prompt: {required_args['prompt']}")

        generation = Generation(
            timestamp=datetime.now().strftime("%Y%m%d_%H%M%S"),
            required_args=required_args,
            optional_args=kwargs
        )
        print('-->generation object done', generation)


        try:
            payload = self._prepare_payload(required_args, **kwargs)
            print('-->payload done')

            # Call Google Veo API
            config = types.GenerateVideosConfig(**kwargs)

            operation = self.client.models.generate_videos(
                model=MODEL_NAME,
                prompt = payload['prompt'],
                image= payload['image'],
                config=config
            )
            print("--> Operation creation done", operation)

            print(f"--> Generation started with operation: {operation.name}")

            # https://ai.google.dev/gemini-api/docs/video#generate-from-images
            # https://github.com/googleapis/python-genai/blob/main/google/genai/operations.py#L348
            # ASYNC FUTURE IMPLEMENTATION
            # # Store operation for later status checking
            # self.operations[operation.name] = {
            #     'operation': operation,
            #     'generation': generation,
            #     'started_at': datetime.now()
            # }

            # print(f"Generation started with operation ID: {operation.name}")
            # generation.update(
            #     call_id = operation,
            #     status=JobStatus.running
            # )
            # generation.to_dict()

            # # Return immediately with operation ID
            # return {
            #     'call_id': operation.name,
            #     'feature_type': 'image_to_video',
            # }

            # Wait for completion
            while not operation.done:
                print("Waiting for video generation to complete...")
                time.sleep(10)
                operation = self.client.operations.get(operation)
            
            # Download the video
            generated_video = operation.response.generated_videos[0]
            video_bytes = self.client.files.download(file=generated_video.video)
            # generated_video.video.save("veo-2-i2v.mp4")
            timestamp = create_timestamp()

            video_filename = f"veo_output_{timestamp}.mp4"

            # Get the directory of this file and save to the same directory
            current_dir = os.path.dirname(os.path.abspath(__file__))
            repo_root = os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(current_dir))))
            output_videos_dir = os.path.join(repo_root, "output_videos")
            os.makedirs(output_videos_dir, exist_ok=True)
            video_path = os.path.join(output_videos_dir, video_filename)

            with open(video_path, 'wb') as f:
                f.write(video_bytes)
            print(f"File downloaded as {video_path}")
            

            generation.update(
                call_id=operation.name,
                status="completed",
                message=f"Video generated and saved to {video_path}",
                result_video=video_path
            )

            return generation.to_dict()

        except Exception as e:
            print(f"Error in generate: {str(e)}")
            generation.update(message=f"Error in generate: {str(e)}")

            raise GenerationError(app_name=APP_NAME, 
                                  model=MODEL_NAME,
                                  feature=FeatureType.IMAGE_TO_VIDEO,
                                  reason=str(e)
                                  )
    # def check_generation_status(self, generation: Generation) -> Generation:
    #     """Check the status of a previous generate call.

    #     Args:
    #         generation: The generation object containing call_id

    #     Returns:
    #         Updated generation object with status and results
    #     """
    #     print('CHECK GEN STATUS')
    #     try:
    #         # where generation['call_id'] is the operation object itself
    #         test_operation = types.Operation(name=generation['call_id'])
    #         operation = self.client.operations.get(test_operation)
    #     except Exception as e:
    #         print('operation not found: ', str(e))

    #     if operation.done:
    #         print('opereation done')
    #         try:
    #             from datetime import datetime
    #             results = []

    #             # Download generated videos
    #             for n, generated_video in enumerate(operation.response.generated_videos):
    #                 # Download video
    #                 self.client.files.download(file=generated_video.video)

    #                 # Save video to file
    #                 timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
    #                 video_filename = f"veo_output_{timestamp}_{n}.mp4"

    #                 # Setup output path
    #                 current_dir = os.path.dirname(os.path.abspath(__file__))
    #                 repo_root = os.path.dirname(os.path.dirname(os.path.dirname(current_dir)))
    #                 output_videos_dir = os.path.join(repo_root, "output_videos")
    #                 os.makedirs(output_videos_dir, exist_ok=True)
    #                 video_path = os.path.join(output_videos_dir, video_filename)

    #                 # Save video
    #                 generated_video.video.save(video_path)

    #                 results.append({
    #                     'filename': video_filename,
    #                     'path': video_path
    #                 })
    #             generation.update(status=JobStatus.COMPLETED, message='Video generation completed', result_video=results)
    #         except Exception as e:
    #             print("Error downloading veo2 generated video, ", str(e))
    #     else:
    #         print("still waiting")
    #         generation.update(timestamp=create_timestamp(), message='Video generation in progress')

registry.register(
    feature="image_to_video",
    model=MODEL_NAME,
    provider="local",
    implementation=Veo20Generate001
)

def base64_to_bytes(data_url):
    if data_url.startswith("data:"):
        header, b64data = data_url.split(",", 1)
    else:
        b64data = data_url
    return base64.b64decode(b64data)

# veo-2.0-generate-001
# veo-3.0-generate-preview
# print("Local generation")
# operation = client.models.generate_videos(
#     model="veo-3.0-generate-preview",
#     prompt="Panning wide shot of a purring kitten sleeping in the sunshine",
#     config=types.GenerateVideosConfig(
#         person_generation="allow_all",  # "allow_adult" and "dont_allow" for Veo 2 only
#         aspect_ratio="16:9",  # "16:9", and "9:16" for Veo 2 only
#     ),
# )

# while not operation.done:
#     time.sleep(20)
#     operation = client.operations.get(operation)

# for n, generated_video in enumerate(operation.response.generated_videos):
#     client.files.download(file=generated_video.video)
#     generated_video.video.save(f"video{n}.mp4")





from fastapi import Body
from fastapi.responses import JSONResponse
import modal
from pathlib import Path

from .base import GPUType, GenericWebAPI, ModalProviderBase
from typing import Any, Dict
from ...core.registry import registry
from ...core._utils import create_timestamp
from ...core.constants import FeatureType
from ...core.errors import (
    APIError,
    DeploymentError
)

"""Configuration Section.
    - Defines
        - Modal app name
        - model ID from HuggingFace
        - cache path to store model data on Modal Volume
        - outputs path to store video outputs on Modal Volume
        - GPU requirements
"""
VOLUME_NAME = "wan-2.1-t2v-14b-cache"
CACHE_PATH = "/cache" 
MODEL_ID = "Wan-AI/Wan2.1-T2V-14B-Diffusers"
OUTPUTS_NAME = "wan-2.1-t2v-14b-outputs"
OUTPUTS_PATH = "/outputs"
APP_NAME = 'wan-2.1-text-to-video-14b'

GPU_CONFIG: GPUType = GPUType.A100_80GB
TIMEOUT: int = 1800 # 30 minutes
SCALEDOWN_WINDOW: int = 900 # 15 minutes

"""Modal Image Definition.
    - Creates a container image with all required depencies (requirements.txt)
    - This image is cached and reused across function calls
"""
image = (
    modal.Image.debian_slim(python_version="3.11")
    .apt_install("git")
    .pip_install(
        "git+https://github.com/huggingface/diffusers.git",
        "torch>=2.4.0",
        "torchvision>=0.19.0",
        "opencv-python>=4.9.0.80",
        "diffusers>=0.31.0",
        "transformers>=4.49.0",
        "tokenizers>=0.20.3",
        "accelerate>=1.1.1",
        "tqdm",
        "imageio",
        "easydict",
        "ftfy",
        "dashscope",
        "imageio-ffmpeg",
        "gradio>=5.0.0",
        "numpy>=1.23.5,<2",
        "fastapi[standard]",
        "peft",
    ).env({"HF_HUB_CACHE": CACHE_PATH})
)

cache_volume = modal.Volume.from_name(VOLUME_NAME, create_if_missing=True)
outputs_volume = modal.Volume.from_name(OUTPUTS_NAME, create_if_missing=True)

# Create Modal app at module level
app = modal.App(APP_NAME)

@app.cls(
    image=image,
    gpu=GPU_CONFIG,
    secrets=[modal.Secret.from_name("huggingface-secret")],
    volumes={CACHE_PATH: cache_volume, OUTPUTS_PATH: outputs_volume},
    timeout=TIMEOUT,
    scaledown_window=SCALEDOWN_WINDOW,
)
class T2V:
    """Modal app implementation
    """
    @modal.enter()
    def load_models(self):
        """This method is called once when the container starts.
        It downloads and initializes the models and pipeline.
        """
        import torch
        from diffusers import AutoModel, WanPipeline
        from diffusers.hooks.group_offloading import apply_group_offloading
        from transformers import UMT5EncoderModel
        import time

        print("Loading models...")
        start_time = time.time()
        print(f"✅ {time.time() - start_time:.2f}s: Starting model load...")

        # Load the models
        try:
            print(f"✅ {time.time() - start_time:.2f}s: Loading text_encoder...")

            text_encoder = UMT5EncoderModel.from_pretrained(
                "Wan-AI/Wan2.1-T2V-14B-Diffusers",
                subfolder="text_encoder",
                torch_dtype=torch.bfloat16,
            )            
            print(f"✅ {time.time() - start_time:.2f}s: text_encoder loaded.")

            print(f"✅ {time.time() - start_time:.2f}s: Loading VAE...")
            vae = AutoModel.from_pretrained(
                "Wan-AI/Wan2.1-T2V-14B-Diffusers",
                subfolder="vae",
                torch_dtype=torch.float32,
            )
            # vae.to("cuda")
            print(f"✅ {time.time() - start_time:.2f}s: VAE loaded.")

            print(f"✅ {time.time() - start_time:.2f}s: Loading transformer...")
            transformer = AutoModel.from_pretrained(
                "Wan-AI/Wan2.1-T2V-14B-Diffusers",
                subfolder="transformer",
                torch_dtype=torch.bfloat16,
            )            
            print(f"✅ {time.time() - start_time:.2f}s: Transformer loaded.")


            print(f"✅ {time.time() - start_time:.2f}s: Creating pipeline...")
            # Apply group-offloading for memory efficiency
            onload_device = torch.device("cuda")
            offload_device = torch.device("cpu")
            apply_group_offloading(
                text_encoder,
                onload_device=onload_device,
                offload_device=offload_device,
                offload_type="block_level",
                num_blocks_per_group=4,
            )
            transformer.enable_group_offload(
                onload_device=onload_device,
                offload_device=offload_device,
                offload_type="leaf_level",
                use_stream=True,
            )

            # Create and run the pipeline
            self.pipeline = WanPipeline.from_pretrained(
                "Wan-AI/Wan2.1-T2V-14B-Diffusers",
                vae=vae,
                transformer=transformer,
                text_encoder=text_encoder,
                torch_dtype=torch.bfloat16,
            )
                        # Completely remove any LoRA to ensure clean base model
            if hasattr(self.pipeline, 'unload_lora_weights'):
                print("unloading lora weights")
                self.pipeline.unload_lora_weights()

            self.pipeline.to("cuda")
            # self.pipeline.enable_model_cpu_offload()
            print(f"✅ {time.time() - start_time:.2f}s: Pipeline ready. Models loaded successfully.")

        except Exception as e:
            raise APIError(response_body=str(e))
        print("Models loaded successfully.")

    @modal.method()
    def generate(self, data: Dict[str, Any]):
        """This method runs the text-to-video generation on an existing container.
        Executes the actual video generating using the loaded pipeline.
        Saves the generated video to Modal's volume storage.
        Returns the video as bytes.
        """
        from diffusers.utils import export_to_video

        print("Generating video...")
        output = self.pipeline(**data).frames[0]

        timestamp = create_timestamp()
        mp4_name = f"wan2.1-pipeline-output_{timestamp}.mp4"

        mp4_path = Path(OUTPUTS_PATH) / mp4_name
        export_to_video(output, mp4_path, fps=16)
        outputs_volume.commit()

        with open(mp4_path, "rb") as f:
            video_bytes = f.read()

        return video_bytes
    @staticmethod
    def handle_web_inference(data: dict):
        """Creates a Modal instance and spawns async job
        Returns immediately with call_id, stored for user"""
        # Create T2V instance and call generate
        try:
            t2v_instance = T2V()
            call = t2v_instance.generate.spawn(data)
        except Exception as e:
            raise DeploymentError(
                service="modal",
                reason=f"Failed to spawn T2V job: {str(e)}",
                app_name=APP_NAME
            )

        return JSONResponse({"call_id": call.object_id, "feature_type": FeatureType.TEXT_TO_VIDEO})
    

class Wan21TextToVideo14B(ModalProviderBase):
    def __init__(self, api_key=None):
        """Initializes the local client that will communicate with Modal.
        Sets references to the Modal app and class names
        """
        super().__init__(api_key)
        self.app_name = APP_NAME
        self.modal_app = app
        self.modal_class_name = "T2V"
    def _prepare_payload(self, required_args, **kwargs) -> Dict[str, Any]:
        """Prepares the data to send to Modal, specific to this model.
        Adds the feature_type field for routing.
        Inherits base validation from parent class
        """
        payload = super()._prepare_payload(required_args, **kwargs)

        # Add feature_type for routing
        payload["feature_type"] = FeatureType.TEXT_TO_VIDEO

        return payload

@app.cls(
    image=image,
    gpu=GPU_CONFIG,
    secrets=[modal.Secret.from_name("huggingface-secret")],
    volumes={CACHE_PATH: cache_volume, OUTPUTS_PATH: outputs_volume},
    timeout=TIMEOUT,
    scaledown_window=SCALEDOWN_WINDOW,
)
class FusionXT2V:
    @modal.enter()
    def load_models(self):
        import torch        
        from diffusers import AutoModel, WanPipeline
        from diffusers.hooks.group_offloading import apply_group_offloading
        from transformers import UMT5EncoderModel

        print("Loading models into GPU memory...")
        text_encoder = UMT5EncoderModel.from_pretrained(
                "Wan-AI/Wan2.1-T2V-14B-Diffusers",
                subfolder="text_encoder",
                torch_dtype=torch.bfloat16,
            )  
        vae = AutoModel.from_pretrained(
                "Wan-AI/Wan2.1-T2V-14B-Diffusers",
                subfolder="vae",
                torch_dtype=torch.bfloat16,
            )
        transformer = AutoModel.from_pretrained(
                "Wan-AI/Wan2.1-T2V-14B-Diffusers",
                subfolder="transformer",
                torch_dtype=torch.bfloat16,
            )
        onload_device = torch.device("cuda")
        offload_device = torch.device("cpu")
        apply_group_offloading(
                text_encoder,
                onload_device=onload_device,
                offload_device=offload_device,
                offload_type="block_level",
                num_blocks_per_group=4,
            )
        transformer.enable_group_offload(
                onload_device=onload_device,
                offload_device=offload_device,
                offload_type="leaf_level",
                use_stream=True,
            )

            # Create and run the pipeline
        self.pipeline = WanPipeline.from_pretrained(
                "Wan-AI/Wan2.1-T2V-14B-Diffusers",
                vae=vae,
                transformer=transformer,
                text_encoder=text_encoder,
                torch_dtype=torch.bfloat16,
            )
        self.pipeline.vae.to(device="cuda")
        print("loading lora weights, enabling")
        # Load T2V FusionX
        self.pipeline.load_lora_weights("vrgamedevgirl84/Wan14BT2VFusioniX", weight_name="FusionX_LoRa/Wan2.1_T2V_14B_FusionX_LoRA.safetensors", adapter_names="T2V FusionX")
        self.pipeline.enable_lora()

        # self.pipeline.to("cuda")
        # self.pipeline.enable_model_cpu_offload()   

        print("✅ Models loaded successfully.")
    @modal.method()
    def generate(self, data: Dict[str, Any]):
        from diffusers.utils import export_to_video
        # for every value in data, pass into pipe
        frames = self.pipeline(**data).frames[0]

        timestamp = create_timestamp()
        mp4_name = f"Wan2.1-T2V-14B-Diffusers-fusionx-t2v-output_{timestamp}.mp4"
        mp4_path = Path(OUTPUTS_PATH) / mp4_name
        export_to_video(frames, str(mp4_path), fps=16)
        outputs_volume.commit()

        with open(mp4_path, "rb") as f:
            video_bytes = f.read()

        return video_bytes 
    @staticmethod
    def handle_web_inference(data: dict):
        """Handle text-to-video generation."""
        # Create T2V instance and call generate
        try:
            t2v_instance = FusionXT2V()
            call = t2v_instance.generate.spawn(data)
        except Exception as e:
            raise DeploymentError(
                service="modal",
                reason=f"Failed to spawn T2V job: {str(e)}",
                app_name=APP_NAME
            )

        return JSONResponse({"call_id": call.object_id, "feature_type": FeatureType.TEXT_TO_VIDEO})
    
class Wan2114bTextToVideoFusionX(ModalProviderBase):
    def __init__(self, api_key=None):
        super().__init__(api_key)
        self.app_name = APP_NAME
        self.modal_app = app
        self.modal_class_name = "FusionXT2V"
    def _prepare_payload(self, required_args, **kwargs) -> Dict[str, Any]:
        """Prepare payload specific to Wan FusionX model."""
        payload = super()._prepare_payload(required_args, **kwargs)

        # Add feature_type for routing
        payload["feature_type"] = FeatureType.TEXT_TO_VIDEO
        payload["model"] = "wan2.1-14b-t2v-fusionx"

        return payload

class WebAPI(GenericWebAPI):
    """Establishes features this model handles"""
    feature_handlers = {
        FeatureType.TEXT_TO_VIDEO: T2V
    }
    @modal.fastapi_endpoint(method="POST")
    def web_inference(self, data: dict = Body(...)):
        """Only implemented for models with LoRAs.
        Routing to LoRA-variant of model
        """

        feature_type = data.pop("feature_type", None)
        model = data.pop("model", None)  # Extract model from data
        if not feature_type:
            return {"error": f"A 'feature_type' is required. Must be one of: {list(self.feature_handlers.keys())}"}

        if feature_type not in self.feature_handlers:
            return {"error": f"Unknown feature_type: {feature_type}. Must be one of: {list(self.feature_handlers.keys())}"}

        # Route to appropriate class based on model and feature_type
        if feature_type == FeatureType.TEXT_TO_VIDEO and model == "wan2.1-14b-t2v-fusionx":
            # Route to FusionT2V for the fusionx model
            handler_class = FusionXT2V
        else:
            # Route to default handler
            handler_class = self.feature_handlers[feature_type]
       
        # If user specifies gpu, timeout, scaledown window, applied here
        if 'modal_options' in data and data['modal_options']:
            modal_options = data['modal_options']
            for key, value in modal_options.items():
                if value is not None:
                    print(f"Applying modal option: {key} = {value}")
                    handler_class = handler_class.with_options(**{key: value})
            data.pop('modal_options')
        return handler_class.handle_web_inference(data)

"""Transforms the WebAPI class into a Modal-compatible class"""
WebAPI = app.cls(
    image=image,
    gpu=GPU_CONFIG,
    secrets=[modal.Secret.from_name("huggingface-secret")],
    volumes={CACHE_PATH: cache_volume, OUTPUTS_PATH: outputs_volume},
    timeout=TIMEOUT,
    scaledown_window=SCALEDOWN_WINDOW,
)(WebAPI)

"""Links the feature/model/provider combination to the implementation class.
Essential to user access to model and its feature(s)
"""
registry.register(
    feature=FeatureType.TEXT_TO_VIDEO,
    model="wan2.1-t2v-14b",
    provider="modal",
    implementation=Wan21TextToVideo14B
)
registry.register(
    feature=FeatureType.TEXT_TO_VIDEO,
    model="wan2.1-14b-t2v-fusionx",
    provider="modal",
    implementation=Wan2114bTextToVideoFusionX,
)


"""Standardized feature types for TioMagic
Defines all supported video generation features
"""
from enum import Enum
from typing import List, Optional
from dataclasses import dataclass, asdict

class FeatureType(str, Enum):
    """Standardized feature types for video generation"""
    TEXT_TO_VIDEO = "text_to_video"
    IMAGE_TO_VIDEO = "image_to_video"
    INTERPOLATE = "interpolate"
    POSE_GUIDANCE = "pose_guidance"

# List of all supported feature types
SUPPORTED_FEATURE_TYPES: List[str] = [
    FeatureType.TEXT_TO_VIDEO,
    FeatureType.IMAGE_TO_VIDEO,
    FeatureType.INTERPOLATE,
    FeatureType.POSE_GUIDANCE,
]

def is_valid_feature_type(feature_type: str) -> bool:
    """Check if a feature type is valid"""
    return feature_type in SUPPORTED_FEATURE_TYPES

def get_feature_types() -> List[str]:
    """Get list of all supported feature types"""
    return SUPPORTED_FEATURE_TYPES.copy()

"""MODAL VIDEO GENERATION TRACKING"""
@dataclass
class Generation:
    call_id: Optional[str] = None
    status: Optional[str] = None
    message: str = ''
    timestamp: Optional[str] = None
    required_args: Optional[dict] = None
    optional_args: Optional[dict] = None
    result_video: Optional[str] = None

    def to_dict(self):
        return asdict(self)

    def update(self, **kwargs):
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)



import modal
from pathlib import Path
from fastapi.responses import JSONResponse

from .base import GPUType, GenericWebAPI, ModalProviderBase
from typing import Any, Dict
from ...core.registry import registry
from ...core._utils import load_image_robust, is_local_path, local_image_to_base64, create_timestamp, extract_image_dimensions
from ...core.constants import FeatureType
from ...core.errors import (
    DeploymentError, ProcessingError
)

APP_NAME = "framepack-i2v-hy"
CACHE_NAME = f"{APP_NAME}-cache"
CACHE_PATH = "/cache"
OUTPUTS_NAME = f'{APP_NAME}-outputs'
OUTPUTS_PATH = "/outputs"

MODEL_ID = "hunyuanvideo-community/HunyuanVideo"

GPU_CONFIG: GPUType = GPUType.A100_80GB
TIMEOUT: int = 3000 # 50 minutes
SCALEDOWN_WINDOW: int = 900 # 15 minutes

image = (
    modal.Image.debian_slim(python_version="3.10")
    .apt_install("git")
    .pip_install(
        "git+https://github.com/huggingface/diffusers.git",
        "torch==2.4", 
        "transformers", 
        "accelerate", 
        "safetensors",
        "sentencepiece", 
        "av", 
        "einops", 
        "Pillow", 
        "numpy",
        "imageio[ffmpeg]",
        "fastapi[standard]",
    )
)

cache_volume = modal.Volume.from_name(CACHE_NAME, create_if_missing=True)
outputs_volume = modal.Volume.from_name(OUTPUTS_NAME, create_if_missing=True)

app = modal.App(APP_NAME)

@app.cls(
    image=image,
    gpu=GPU_CONFIG,
    secrets=[modal.Secret.from_name("huggingface-secret")],
    volumes={CACHE_PATH: cache_volume, OUTPUTS_PATH: outputs_volume},
    timeout=TIMEOUT,
    scaledown_window=SCALEDOWN_WINDOW,
)
class I2V:
    @modal.enter()
    def load_models(self):
        import torch
        from diffusers import HunyuanVideoFramepackPipeline, HunyuanVideoFramepackTransformer3DModel
        from transformers import SiglipImageProcessor, SiglipVisionModel

        transformer = HunyuanVideoFramepackTransformer3DModel.from_pretrained(
            "lllyasviel/FramePackI2V_HY", torch_dtype=torch.bfloat16
        )
        feature_extractor = SiglipImageProcessor.from_pretrained(
            "lllyasviel/flux_redux_bfl", subfolder="feature_extractor"
        )
        image_encoder = SiglipVisionModel.from_pretrained(
            "lllyasviel/flux_redux_bfl", subfolder="image_encoder", torch_dtype=torch.float16
        )
        self.pipe = HunyuanVideoFramepackPipeline.from_pretrained(
            "hunyuanvideo-community/HunyuanVideo",
            transformer=transformer,
            feature_extractor=feature_extractor,
            image_encoder=image_encoder,
            torch_dtype=torch.float16,
        )
        self.pipe.to("cuda")

        self.pipe.vae.enable_tiling()
        print("✅ Complete Framepack pipeline loaded and ready.")

    @modal.method()
    def generate(self, data: Dict[str, Any]):
        from diffusers.utils import export_to_video
        frames = self.pipe(
            **data,
            sampling_type="inverted_anti_drifting").frames[0]

        timestamp = create_timestamp()
        mp4_name = f"framepack-output_{timestamp}.mp4"
        mp4_path = Path(OUTPUTS_PATH) / mp4_name
        export_to_video(frames, str(mp4_path), fps=16)
        outputs_volume.commit()

        with open(mp4_path, "rb") as f:
            video_bytes = f.read()

        return video_bytes 
    @staticmethod
    def handle_web_inference(data: Dict[str, Any]): 
        image = data.get("image")

        try:
            image = load_image_robust(image)
            data['image'] = image
            if 'height' not in data and 'width' not in data:
                data = extract_image_dimensions(image, data)
        except Exception as e:
            raise ProcessingError(
                media_type="image",
                operation="load and process",
                reason=str(e),
                file_path=data.get("image") if isinstance(data.get("image"), str) else None
            )

        # Create I2V instance and call generate
        try:
            i2v_instance = I2V()
            call = i2v_instance.generate.spawn(data)
        except Exception as e:
            raise DeploymentError(
                service="Modal",
                reason=f"Failed to spawn I2V job: {str(e)}",
                app_name=APP_NAME
            )

        return JSONResponse({"call_id": call.object_id, "feature_type": FeatureType.IMAGE_TO_VIDEO})
class FramepackI2VHYImageToVideo(ModalProviderBase):
    def __init__(self, api_key=None):
        super().__init__(api_key)
        self.app_name = APP_NAME
        self.modal_app = app
        self.modal_class_name = "I2V"
    def _prepare_payload(self, required_args: Dict[str, Any], **kwargs) -> Dict[str, Any]:
        """Prepare payload specific to Wan2.1 Vace Image-to-Video model."""
        payload = super()._prepare_payload(required_args, **kwargs)
        payload["feature_type"] = FeatureType.IMAGE_TO_VIDEO

        if is_local_path(payload['image']):
            payload['image'] = local_image_to_base64(payload['image'])
        return payload

@app.cls(
    image=image,
    gpu=GPU_CONFIG,
    secrets=[modal.Secret.from_name("huggingface-secret")],
    volumes={CACHE_PATH: cache_volume, OUTPUTS_PATH: outputs_volume},
    timeout=TIMEOUT,
    scaledown_window=SCALEDOWN_WINDOW,
)
class Interpolate:
    @modal.enter()
    def load_models(self):
        import torch
        from diffusers import HunyuanVideoFramepackPipeline, HunyuanVideoFramepackTransformer3DModel
        from transformers import SiglipImageProcessor, SiglipVisionModel

        transformer = HunyuanVideoFramepackTransformer3DModel.from_pretrained(
            "lllyasviel/FramePackI2V_HY", torch_dtype=torch.bfloat16
        )
        feature_extractor = SiglipImageProcessor.from_pretrained(
            "lllyasviel/flux_redux_bfl", subfolder="feature_extractor"
        )
        image_encoder = SiglipVisionModel.from_pretrained(
            "lllyasviel/flux_redux_bfl", subfolder="image_encoder", torch_dtype=torch.float16
        )
        self.pipe = HunyuanVideoFramepackPipeline.from_pretrained(
            "hunyuanvideo-community/HunyuanVideo",
            transformer=transformer,
            feature_extractor=feature_extractor,
            image_encoder=image_encoder,
            torch_dtype=torch.float16,
        )

        self.pipe.to("cuda")
        self.pipe.enable_model_cpu_offload()
        self.pipe.vae.enable_tiling()
        print("✅ Complete Framepack pipeline loaded and ready.")
    @modal.method()
    def generate(self, data: Dict[str, Any]):
        from diffusers.utils import export_to_video

        # adjust toolkit requirements to match framepack params
        data['image'] = data.pop('first_frame')
        data['last_image'] = data.pop('last_frame')

        frames = self.pipe(
            **data,
            sampling_type="inverted_anti_drifting").frames[0]

        timestamp = create_timestamp()
        mp4_name = f"framepack-output_{timestamp}.mp4"
        mp4_path = Path(OUTPUTS_PATH) / mp4_name
        export_to_video(frames, str(mp4_path), fps=16)
        outputs_volume.commit()

        with open(mp4_path, "rb") as f:
            video_bytes = f.read()

        return video_bytes 
    @staticmethod
    def handle_web_inference(data: Dict[str, Any]):
        first_frame = data.get("first_frame")
        last_frame = data.get("last_frame")

        try:
            # load_image_robust can handle both URLs and base64 strings
            first_frame = load_image_robust(first_frame)
            last_frame = load_image_robust(last_frame)   
            data['first_frame'] = first_frame
            data['last_frame'] = last_frame    
            if 'height' not in data and 'width' not in data:
                data = extract_image_dimensions(first_frame, data)     
        except Exception as e:
            raise ProcessingError(
                media_type="image",
                operation="load and process",
                reason=str(e),
                file_path=data.get("image") if isinstance(data.get("image"), str) else None
            )

        try:
            interpolate_instance = Interpolate()
            call = interpolate_instance.generate.spawn(data)
        except Exception as e:
            raise DeploymentError(
                service="Modal",
                reason=f"Failed to spawn Interpolate job: {str(e)}",
                app_name=APP_NAME
            )

        return JSONResponse({"call_id": call.object_id, "feature_type": FeatureType.INTERPOLATE})
class FramepackI2VHYInterpolate(ModalProviderBase):
    def __init__(self, api_key=None):
        super().__init__(api_key)
        self.app_name = APP_NAME
        self.modal_app = app
        self.modal_class_name = "Interpolate"
    def _prepare_payload(self, required_args: Dict[str, Any], **kwargs) -> Dict[str, Any]:
        """Prepare payload specific to Framepack I2V HY Interpolate model.
        Break out required args into payload
        """
        payload = super()._prepare_payload(required_args, **kwargs)
        # payload = {"prompt": required_args['prompt']}
        payload["feature_type"] = FeatureType.INTERPOLATE

        if payload['first_frame'] is None or payload['last_frame'] is None:
            raise ValueError("Arguments 'first_frame' and 'last_frame' are required for Interpolation Video generation")

        if is_local_path(payload['first_frame']):
            # Convert local image to base64
            payload["first_frame"] = local_image_to_base64(payload['first_frame'])
        if is_local_path(payload['last_frame']):
            # Convert local image to base64
            payload["last_frame"] = local_image_to_base64(payload['last_frame'])

        return payload

# Create a subclass with the handlers
class WebAPI(GenericWebAPI):
    feature_handlers = {
        FeatureType.INTERPOLATE: Interpolate,
        FeatureType.IMAGE_TO_VIDEO: I2V,
    }

# Apply Modal decorator
WebAPI = app.cls(
    image=image,
    gpu=GPU_CONFIG,
    secrets=[modal.Secret.from_name("huggingface-secret")],
    volumes={CACHE_PATH: cache_volume, OUTPUTS_PATH: outputs_volume},
    timeout=TIMEOUT,
    scaledown_window=SCALEDOWN_WINDOW,
)(WebAPI)

registry.register(
    feature=FeatureType.INTERPOLATE,
    model="framepack-i2v-hy",
    provider="modal",
    implementation=FramepackI2VHYInterpolate
)
registry.register(
    feature=FeatureType.IMAGE_TO_VIDEO,
    model="framepack-i2v-hy",
    provider="modal",
    implementation=FramepackI2VHYImageToVideo
)





import os
from pathlib import Path
import base64
from typing import Any, Dict
from ..errors import ProcessingError


def is_local_path(path: str) -> bool:
    """Check if the path is a local file path."""
    if not isinstance(path, str):
        return False

    # Check if it's a URL
    if path.startswith(('http://', 'https://')):
        return False

    # Check if it's a local file path
    return os.path.exists(path) or Path(path).exists()

def local_image_to_base64(image_path: str) -> str:
    """Convert local image to base64 string."""
    import base64

    try:
        with open(image_path, "rb") as f:
            image_data = f.read()

        # Convert to base64
        base64_str = base64.b64encode(image_data).decode('utf-8')

        # Add data URL prefix for image type detection
        # You might want to detect the actual image type here
        return f"data:image/png;base64,{base64_str}"

    except Exception as e:
        # raise Exception(f"Failed to convert local image {image_path} to base64: {e}")
        raise ProcessingError(operation = "convert local image to base64", reason=str(e), media_type="image")

def extract_image_dimensions(image, data: Dict[str, Any]) -> Dict[str, Any]:
    """get image dimensions and set them in data"""

    # Handle different types of image objects
    if hasattr(image, 'size'):
        # PIL Image or similar
        width, height = image.size
    elif hasattr(image, 'width') and hasattr(image, 'height'):
        # Google GenAI Image or similar
        width, height = image.width, image.height
    else:
        # Try to get dimensions from other attributes
        try:
            # For Google GenAI Image objects, they might have different attributes
            if hasattr(image, '_image'):
                # Try to access the underlying image
                pil_image = image._image
                width, height = pil_image.size
            else:
                print(f"Warning: Could not extract dimensions from image object of type {type(image)}")
                return data
        except Exception as e:
            print(f"Warning: Could not extract dimensions from image: {e}")
            return data
    
    data['width'] = width
    data['height'] = height
    print(f"Extracted image dimensions: {width}x{height}")
    return data



def url_video_to_bytes(url: str) -> bytes:
    import requests
    response = requests.get(url)
    response.raise_for_status()
    return response.content

def load_video_robust(video_source):
    try:
        # check if web url
        if isinstance(video_source, str):
            if video_source.startswith(('http://', 'https://')):
                print(f"Loading video from URL: {video_source}")
                return url_video_to_bytes(video_source)
            elif video_source.startswith('data:video'):
                print("Loading video from base64 data URL")
                # Extract base64 part from data URL
                if "," in video_source:
                    base64_str = video_source.split(",")[1]
                else:
                    base64_str = video_source
                return base64.b64decode(base64_str)
            # Check if it's a raw base64 string (without data URL prefix)
            elif len(video_source) > 100:  # Base64 strings are typically long
                try:
                    print("Loading video from base64 string")
                    return base64.b64decode(video_source)
                except Exception:
                    # If base64 decode fails, treat as local path
                    pass
        # If it's already bytes, return as-is
        elif isinstance(video_source, bytes):
            return video_source
        else:
            raise ValueError(f"Unsupported video source format: {type(video_source)}")
    except Exception as e:
        print(f"Error loading video from {video_source}: {e}")
        raise ProcessingError(media_type="video", operation="load video", reason=str(e))

def create_timestamp():
    from datetime import datetime
    return datetime.now().strftime("%m%d_%H%M%S")


"""Exception hierarchy for TioMagic Animation API.

This module defines custom exceptions for handling various error scenarios
in the TioMagic animation system.
"""

from typing import Optional, Dict, Any


class TioMagicError(Exception):
    """Base exception for all TioMagic errors."""
    
    def __init__(self, message: str, details: Optional[Dict[str, Any]] = None):
        super().__init__(message)
        self.message = message
        self.details = details or {}
        
    def to_dict(self) -> Dict[str, Any]:
        """Convert exception to dictionary for API responses."""
        return {
            "error": self.__class__.__name__,
            "message": self.message,
            "details": self.details
        }


class UnknownModelError(TioMagicError):
    """Raised when a model string can't be found in the registry."""
    
    def __init__(self, model: str, available_models: Optional[list] = None):
        message = f"Unknown model: '{model}'"
        if available_models:
            message += f". Available models: {', '.join(available_models)}"
        super().__init__(message, {"model": model, "available_models": available_models})
        self.model = model
        self.available_models = available_models


class UnknownProviderError(TioMagicError):
    """Raised when a provider string can't be found."""
    
    def __init__(self, provider: str, available_providers: Optional[list] = None):
        message = f"Unknown provider: '{provider}'"
        if available_providers:
            message += f". Available providers: {', '.join(available_providers)}"
        super().__init__(message, {"provider": provider, "available_providers": available_providers})
        self.provider = provider
        self.available_providers = available_providers


class JobTimeoutError(TioMagicError):
    """Raised when Job.wait() exceeds time limit before the job finishes."""
    
    def __init__(self, job_id: str, timeout_seconds: int):
        message = f"Job '{job_id}' timed out after {timeout_seconds} seconds"
        super().__init__(message, {"job_id": job_id, "timeout_seconds": timeout_seconds})
        self.job_id = job_id
        self.timeout_seconds = timeout_seconds


class JobExecutionError(TioMagicError):
    """Raised when provider reports that job has explicitly failed."""
    
    def __init__(self, job_id: str, reason: str, provider_error: Optional[str] = None):
        message = f"Job '{job_id}' failed: {reason}"
        details = {"job_id": job_id, "reason": reason}
        if provider_error:
            details["provider_error"] = provider_error
        super().__init__(message, details)
        self.job_id = job_id
        self.reason = reason
        self.provider_error = provider_error


class AuthenticationError(TioMagicError):
    """Raised when there's a missing or invalid API key or token."""
    
    def __init__(self, provider: Optional[str] = None, message: Optional[str] = None):
        if message is None:
            message = "Authentication failed"
            if provider:
                message += f" for provider '{provider}'"
        super().__init__(message, {"provider": provider})
        self.provider = provider


class APIError(TioMagicError):
    """Raised for network-related issues, like non-200 responses or JSON parsing failures."""
    
    def __init__(self, message: str, status_code: Optional[int] = None, 
                 response_body: Optional[str] = None, url: Optional[str] = None):
        details = {}
        if status_code:
            details["status_code"] = status_code
        if response_body:
            details["response_body"] = response_body[:500]  # Truncate long responses
        if url:
            details["url"] = url
        super().__init__(message, details)
        self.status_code = status_code
        self.response_body = response_body
        self.url = url


class DeploymentError(TioMagicError):
    """Raised when target infrastructure is unavailable (e.g., Modal app name is wrong)."""
    
    def __init__(self, service: str, reason: str, app_name: Optional[str] = None):
        message = f"Deployment error for {service}: {reason}"
        details = {"service": service, "reason": reason}
        if app_name:
            details["app_name"] = app_name
        super().__init__(message, details)
        self.service = service
        self.reason = reason
        self.app_name = app_name


class ValidationError(TioMagicError):
    """Raised when user provided invalid input parameters."""
    
    def __init__(self, field: str, message: str, value: Any = None, 
                 constraints: Optional[Dict[str, Any]] = None):
        full_message = f"Validation error for '{field}': {message}"
        details = {"field": field, "message": message}
        if value is not None:
            details["provided_value"] = value
        if constraints:
            details["constraints"] = constraints
        super().__init__(full_message, details)
        self.field = field
        self.value = value
        self.constraints = constraints


class RateLimitError(TioMagicError):
    """Raised when provider's rate limit has been exceeded."""
    
    def __init__(self, provider: str, retry_after: Optional[int] = None, 
                 limit_type: Optional[str] = None):
        message = f"Rate limit exceeded for provider '{provider}'"
        details = {"provider": provider}
        if retry_after:
            message += f". Retry after {retry_after} seconds"
            details["retry_after"] = retry_after
        if limit_type:
            details["limit_type"] = limit_type
        super().__init__(message, details)
        self.provider = provider
        self.retry_after = retry_after
        self.limit_type = limit_type


class ResourceNotFoundError(TioMagicError):
    """Raised when a requested resource (image, video, etc.) cannot be found."""
    
    def __init__(self, resource_type: str, resource_id: str, location: Optional[str] = None):
        message = f"{resource_type} '{resource_id}' not found"
        if location:
            message += f" at {location}"
        super().__init__(message, {
            "resource_type": resource_type,
            "resource_id": resource_id,
            "location": location
        })
        self.resource_type = resource_type
        self.resource_id = resource_id
        self.location = location


class ProcessingError(TioMagicError):
    """Raised when there's an error processing media (images, videos)."""
    
    def __init__(self, media_type: str, operation: str, reason: str, 
                 file_path: Optional[str] = None):
        message = f"Failed to {operation} {media_type}: {reason}"
        details = {
            "media_type": media_type,
            "operation": operation,
            "reason": reason
        }
        if file_path:
            details["file_path"] = file_path
        super().__init__(message, details)
        self.media_type = media_type
        self.operation = operation
        self.reason = reason
        self.file_path = file_path

class ConfigurationError(TioMagicError):
    """Raised when there's a configuration issue."""
    
    def __init__(self, config_type: str, message: str, missing_field: Optional[str] = None):
        full_message = f"Configuration error for {config_type}: {message}"
        details = {"config_type": config_type}
        if missing_field:
            details["missing_field"] = missing_field
        super().__init__(full_message, details)
        self.config_type = config_type
        self.missing_field = missing_field


class ImplementationError(TioMagicError):
    """Raised when implementation cannot be instantiated or is missing required methods."""
    
    def __init__(self, provider: str, feature: str, reason: str, 
                 implementation_class: Optional[str] = None):
        message = f"Failed to create {feature} implementation for provider '{provider}': {reason}"
        details = {
            "provider": provider,
            "feature": feature,
            "reason": reason
        }
        if implementation_class:
            details["implementation_class"] = implementation_class
        super().__init__(message, details)
        self.provider = provider
        self.feature = feature
        self.reason = reason

class GenerationError(TioMagicError):
    """Raised when the AI model generation process fails.
    
    This is distinct from JobExecutionError which is about job infrastructure.
    GenerationError specifically captures failures during the actual AI generation
    process (model inference, GPU errors, output generation, etc.)
    """
    
    def __init__(self, 
                 app_name: str,
                 model: str,
                 feature: str,
                 reason: str,
                 generation_params: Optional[Dict[str, Any]] = None,
                 gpu_info: Optional[Dict[str, Any]] = None,
                 traceback: Optional[str] = None):
        message = f"Generation failed for {feature} using model '{model}': {reason}"
        details = {
            "model": model,
            "feature": feature,
            "reason": reason
        }
        
        if generation_params:
            # Don't include sensitive data like full prompts
            safe_params = {
                k: v for k, v in generation_params.items() 
                if k in ["height", "width", "num_frames", "guidance_scale", "num_inference_steps"]
            }
            details["generation_params"] = safe_params
            
        if gpu_info:
            details["gpu_info"] = gpu_info
            
        if traceback:
            # Truncate long tracebacks
            details["traceback"] = traceback[-1000:] if len(traceback) > 1000 else traceback
            
        super().__init__(message, details)
        self.app_name = app_name
        self.model = model
        self.feature = feature
        self.reason = reason
        self.generation_params = generation_params
        self.gpu_info = gpu_info
        self.traceback = traceback